import torch
import torch.nn as nn
import torch.nn.functional as F
import typing as T
from torch_geometric.nn import MLP
from massspecgym.models.tokenizers import SpecialTokensBaseTokenizer
from massspecgym.data.transforms import MolToFormulaVector
from massspecgym.models.base import Stage
from massspecgym.models.de_novo.base import DeNovoMassSpecGymModel
from massspecgym.definitions import PAD_TOKEN, SOS_TOKEN, EOS_TOKEN


class SmilesTransformer(DeNovoMassSpecGymModel):
    def __init__(
        self,
        input_dim: int,
        d_model: int,
        nhead: int,
        num_encoder_layers: int,
        num_decoder_layers: int,
        smiles_tokenizer: SpecialTokensBaseTokenizer,
        start_token: str = SOS_TOKEN,
        end_token: str = EOS_TOKEN,
        pad_token: str = PAD_TOKEN,
        dropout: float = 0.1,
        max_smiles_len: int = 200,
        k_predictions: int = 1,
        temperature: T.Optional[float] = 1.0,
        pre_norm: bool = False,
        chemical_formula: bool = False,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.smiles_tokenizer = smiles_tokenizer
        self.vocab_size = smiles_tokenizer.get_vocab_size()
        for token in [start_token, end_token, pad_token]:
            assert token in smiles_tokenizer.get_vocab(), f"Token {token} not found in tokenizer vocabulary."
        self.start_token_id = smiles_tokenizer.token_to_id(start_token)
        self.end_token_id = smiles_tokenizer.token_to_id(end_token)
        self.pad_token_id = smiles_tokenizer.token_to_id(pad_token)

        self.d_model = d_model
        self.max_smiles_len = max_smiles_len
        self.k_predictions = k_predictions
        self.temperature = temperature
        if self.k_predictions == 1:  # TODO: this logic should be changed because sampling with k = 1 also makes sense
            self.temperature = None

        self.src_encoder = nn.Linear(input_dim, d_model)
        self.tgt_embedding = nn.Embedding(self.vocab_size, d_model)
        self.transformer = nn.Transformer(
            d_model=d_model,
            nhead=nhead,
            num_encoder_layers=num_encoder_layers,
            num_decoder_layers=num_decoder_layers,
            dim_feedforward=4 * d_model,
            dropout=dropout,
            norm_first=pre_norm
        )
        self.tgt_decoder = nn.Linear(d_model, self.vocab_size)

        self.chemical_formula = chemical_formula
        if self.chemical_formula:
            self.formula_mlp = MLP(
                in_channels=MolToFormulaVector.num_elements(),
                hidden_channels=MolToFormulaVector.num_elements(),
                out_channels=d_model,
                num_layers=1,
                dropout=dropout,
                norm=None
            )

        self.criterion = nn.CrossEntropyLoss()

    def forward(self, batch):
        
        spec = batch["spec"]  # (batch_size, seq_len, in_dim)
        smiles = batch["mol"]  # List of SMILES of length batch_size

        smiles = self.smiles_tokenizer.encode_batch(smiles)
        smiles = [s.ids for s in smiles]
        smiles = torch.tensor(smiles, device=spec.device)  # (batch_size, seq_len)

        # Generating padding masks for variable-length sequences
        src_key_padding_mask = self.generate_src_padding_mask(spec)
        tgt_key_padding_mask = self.generate_tgt_padding_mask(smiles)

        # Create target mask (causal mask)
        tgt_seq_len = smiles.size(1)
        tgt_mask = self.transformer.generate_square_subsequent_mask(tgt_seq_len).to(smiles.device)

        # Preapre inputs for transformer teacher forcing
        src = spec.permute(1, 0, 2)  # (seq_len, batch_size, in_dim)
        smiles = smiles.permute(1, 0)  # (seq_len, batch_size)
        tgt = smiles[:-1, :]
        tgt_mask = tgt_mask[:-1, :-1]
        src_key_padding_mask = src_key_padding_mask
        tgt_key_padding_mask = tgt_key_padding_mask[:, :-1]

        # Input and output embeddings
        src = self.src_encoder(src)  # (seq_len, batch_size, d_model)
        if self.chemical_formula:
            formula_emb = self.formula_mlp(batch["formula"])  # (batch_size, d_model)
            src = src + formula_emb.unsqueeze(0)  # (seq_len, batch_size, d_model) + (1, batch_size, d_model)
        src = src * (self.d_model**0.5)
        tgt = self.tgt_embedding(tgt) * (self.d_model**0.5)  # (seq_len, batch_size, d_model)

        # Transformer forward pass
        memory = self.transformer.encoder(src, src_key_padding_mask=src_key_padding_mask)
        output = self.transformer.decoder(tgt, memory, tgt_mask=tgt_mask, tgt_key_padding_mask=tgt_key_padding_mask)

        # Logits to vocabulary
        output = self.tgt_decoder(output)  # (seq_len, batch_size, vocab_size)

        # Reshape before returning
        smiles_pred = output.view(-1, self.vocab_size)
        smiles = smiles[1:, :].contiguous().view(-1)
        return smiles_pred, smiles

    def step(self, batch: dict, stage: Stage = Stage.NONE) -> dict:

        # Forward pass
        smiles_pred, smiles = self.forward(batch)

        # Compute loss
        loss = self.criterion(smiles_pred, smiles)

        # Generate SMILES strings
        if stage in self.log_only_loss_at_stages:
            mols_pred = None
        else:
            mols_pred = self.decode_smiles(batch)

        return dict(loss=loss, mols_pred=mols_pred)

    def generate_src_padding_mask(self, spec):
        return spec.sum(-1) == 0

    def generate_tgt_padding_mask(self, smiles):
        return smiles == self.pad_token_id

    def decode_smiles(self, batch):

        decoded_smiles_str = []
        for _ in range(self.k_predictions):
            decoded_smiles = self.greedy_decode(
                batch,
                max_len=self.max_smiles_len,
                temperature=self.temperature,
            )

            decoded_smiles = [seq.tolist() for seq in decoded_smiles]
            decoded_smiles_str.append(self.smiles_tokenizer.decode_batch(decoded_smiles))

        # Transpose from (k, batch_size) to (batch_size, k)
        decoded_smiles_str = list(map(list, zip(*decoded_smiles_str)))

        return decoded_smiles_str

    def greedy_decode(self, batch, max_len, temperature):

        with torch.inference_mode():

            spec = batch["spec"]    # (batch_size, seq_len, in_dim) 
            src_key_padding_mask = self.generate_src_padding_mask(spec)   

            spec = spec.permute(1, 0, 2)  # (seq_len, batch_size, in_dim)
            src = self.src_encoder(spec)  # (seq_len, batch_size, d_model)
            if self.chemical_formula:
                formula_emb = self.formula_mlp(batch["formula"])  # (batch_size, d_model)
                src = src + formula_emb.unsqueeze(0)  # (seq_len, batch_size, d_model) + (1, batch_size, d_model)
            src = src * (self.d_model**0.5)
            memory = self.transformer.encoder(src, src_key_padding_mask=src_key_padding_mask)

            batch_size = src.size(1)
            out_tokens = torch.ones(1, batch_size).fill_(self.start_token_id).type(torch.long).to(spec.device)

            for _ in range(max_len - 1):
                tgt = self.tgt_embedding(out_tokens) * (self.d_model**0.5)
                tgt_mask = self.transformer.generate_square_subsequent_mask(tgt.size(0)).to(src.device)
                out = self.transformer.decoder(tgt, memory, tgt_mask=tgt_mask)
                out = self.tgt_decoder(out[-1, :])  # (batch_size, vocab_size)

                # Select next token
                if self.temperature is None:
                    probs = F.softmax(out, dim=-1)
                    next_token = torch.argmax(probs, dim=-1)  # (batch_size,)
                else:
                    probs = F.softmax(out / temperature, dim=-1)
                    next_token = torch.multinomial(probs, num_samples=1).squeeze(1)  # (batch_size,)

                next_token = next_token.unsqueeze(0)  # (1, batch_size)

                out_tokens = torch.cat([out_tokens, next_token], dim=0)
                if torch.all(next_token == self.end_token_id):
                    break

            out_tokens = out_tokens.permute(1, 0)  # (batch_size, seq_len)
            return out_tokens  



import torch
import torch.nn as nn
import numpy as np

DEFAULT_MAX_COUNT_INT = 255
DEFAULT_NUM_EXTRA_EMBEDDINGS = 1

class IntFeaturizer(nn.Module):
	"""
	Base class for mapping integers to a vector representation (primarily to be used as a "richer" embedding for NNs
	processing integers).

	Subclasses should define `self.int_to_feat_matrix`, a matrix where each row is the vector representation for that
	integer, i.e. to get a vector representation for `5`, one could call `self.int_to_feat_matrix[5]`.

	Note that this class takes care of creating a fixed number (`self.NUM_EXTRA_EMBEDDINGS` to be precise) of extra
	"learned" embeddings these will be concatenated after the integer embeddings in the forward pass,
	be learned, and be used for extra  non-integer tokens such as the "to be confirmed token" (i.e., pad) token.
	They are indexed starting from `self.max_count_int`.
	"""

	def __init__(self, embedding_dim, max_count_int=DEFAULT_MAX_COUNT_INT, num_extra_embeddings=DEFAULT_NUM_EXTRA_EMBEDDINGS):
		super().__init__()
		self.max_count_int = max_count_int
		self.num_extra_embeddings = num_extra_embeddings
		weights = torch.zeros(self.num_extra_embeddings, embedding_dim)
		self._extra_embeddings = nn.Parameter(weights, requires_grad=True)
		nn.init.normal_(self._extra_embeddings, 0.0, 1.0)
		self.embedding_dim = embedding_dim

	def forward(self, tensor):
		"""
		Convert the integer `tensor` into its new representation -- note that it gets stacked along final dimension.
		"""
		orig_shape = tensor.shape
		# out_tensor = torch.empty(
		# 	(*orig_shape, self.embedding_dim), device=tensor.device
		# )
		extra_embed_mask = (tensor >= self.max_count_int).float()

		tensor = tensor.long()
		norm_embeds = self.int_to_feat_matrix[tensor]
		extra_embeds = self._extra_embeddings[torch.maximum(tensor,self.max_count_int*torch.ones_like(tensor))-self.max_count_int]
		out_tensor = (1.-extra_embed_mask).unsqueeze(-1)*norm_embeds + extra_embed_mask.unsqueeze(-1)*extra_embeds

		temp_out = out_tensor.reshape(*orig_shape[:-1], -1)
		return temp_out

	@property
	def num_dim(self):
		return self.int_to_feat_matrix.shape[1]

	# @property
	# def full_dim(self):
	#	 return self.num_dim * common.NORM_VEC.shape[0]


class FourierFeaturizer(IntFeaturizer):
	"""
	Inspired by:
	Tancik, M., Srinivasan, P.P., Mildenhall, B., Fridovich-Keil, S., Raghavan, N., Singhal, U., Ramamoorthi, R.,
	Barron, J.T. and Ng, R. (2020) ‘Fourier Features Let Networks Learn High Frequency Functions in Low Dimensional
	 Domains’, arXiv [cs.CV]. Available at: http://arxiv.org/abs/2006.10739.

	Some notes:
	* we'll put the frequencies at powers of 1/2 rather than random Gaussian samples; this means it will match the
		Binarizer quite closely but be a bit smoother.
	"""

	def __init__(self, max_count_int=DEFAULT_MAX_COUNT_INT, num_extra_embeddings=DEFAULT_NUM_EXTRA_EMBEDDINGS):

		num_freqs = int(np.ceil(np.log2(max_count_int))) + 2
		# ^ need at least this many to ensure that the whole input range can be represented on the half circle.

		freqs = 0.5 ** torch.arange(num_freqs, dtype=torch.float32)
		freqs_time_2pi = 2 * np.pi * freqs

		super().__init__(
			embedding_dim=2 * freqs_time_2pi.shape[0],
			max_count_int=max_count_int,
			num_extra_embeddings=num_extra_embeddings,
		)  # 2 for cosine and sine

		# we will define the features at this frequency up front (as we only will ever see a fixed number of counts):
		combo_of_sinusoid_args = (
			torch.arange(self.max_count_int, dtype=torch.float32)[:, None]
			* freqs_time_2pi[None, :]
		)
		all_features = torch.cat(
			[torch.cos(combo_of_sinusoid_args), torch.sin(combo_of_sinusoid_args)],
			dim=1,
		)

		# ^ shape:  MAX_COUNT_INT x 2 * num_freqs
		self.int_to_feat_matrix = nn.Parameter(all_features.float())
		self.int_to_feat_matrix.requires_grad = False


class FourierFeaturizerSines(IntFeaturizer):
	"""
	Like other fourier feats but sines only

	Inspired by:
	Tancik, M., Srinivasan, P.P., Mildenhall, B., Fridovich-Keil, S., Raghavan, N., Singhal, U., Ramamoorthi, R.,
	Barron, J.T. and Ng, R. (2020) ‘Fourier Features Let Networks Learn High Frequency Functions in Low Dimensional
	 Domains’, arXiv [cs.CV]. Available at: http://arxiv.org/abs/2006.10739.

	Some notes:
	* we'll put the frequencies at powers of 1/2 rather than random Gaussian samples; this means it will match the
		Binarizer quite closely but be a bit smoother.
	"""

	def __init__(self, max_count_int=DEFAULT_MAX_COUNT_INT, num_extra_embeddings=DEFAULT_NUM_EXTRA_EMBEDDINGS):

		num_freqs = int(np.ceil(np.log2(max_count_int))) + 2
		# ^ need at least this many to ensure that the whole input range can be represented on the half circle.

		freqs = (0.5 ** torch.arange(num_freqs, dtype=torch.float32))[2:]
		freqs_time_2pi = 2 * np.pi * freqs

		super().__init__(
			embedding_dim=freqs_time_2pi.shape[0],
			max_count_int=max_count_int,
			num_extra_embeddings=num_extra_embeddings	
		)

		# we will define the features at this frequency up front (as we only will ever see a fixed number of counts):
		combo_of_sinusoid_args = (
			torch.arange(self.max_count_int, dtype=torch.float32)[:, None]
			* freqs_time_2pi[None, :]
		)
		# ^ shape:  MAX_COUNT_INT x 2 * num_freqs
		self.int_to_feat_matrix = nn.Parameter(
			torch.sin(combo_of_sinusoid_args).float()
		)
		self.int_to_feat_matrix.requires_grad = False


class FourierFeaturizerAbsoluteSines(IntFeaturizer):
	"""
	Like other fourier feats but sines only and absoluted.

	Inspired by:
	Tancik, M., Srinivasan, P.P., Mildenhall, B., Fridovich-Keil, S., Raghavan, N., Singhal, U., Ramamoorthi, R.,
	Barron, J.T. and Ng, R. (2020) ‘Fourier Features Let Networks Learn High Frequency Functions in Low Dimensional
	 Domains’, arXiv [cs.CV]. Available at: http://arxiv.org/abs/2006.10739.

	Some notes:
	* we'll put the frequencies at powers of 1/2 rather than random Gaussian samples; this means it will match the
		Binarizer quite closely but be a bit smoother.
	"""

	def __init__(self, max_count_int=DEFAULT_MAX_COUNT_INT, num_extra_embeddings=DEFAULT_NUM_EXTRA_EMBEDDINGS):

		num_freqs = int(np.ceil(np.log2(max_count_int))) + 2

		freqs = (0.5 ** torch.arange(num_freqs, dtype=torch.float32))[2:]
		freqs_time_2pi = 2 * np.pi * freqs

		super().__init__(
			embedding_dim=freqs_time_2pi.shape[0],
			max_count_int=max_count_int,
			num_extra_embeddings=num_extra_embeddings
		)

		# we will define the features at this frequency up front (as we only will ever see a fixed number of counts):
		combo_of_sinusoid_args = (
			torch.arange(self.max_count_int, dtype=torch.float32)[:, None]
			* freqs_time_2pi[None, :]
		)
		# ^ shape:  MAX_COUNT_INT x 2 * num_freqs
		self.int_to_feat_matrix = nn.Parameter(
			torch.abs(torch.sin(combo_of_sinusoid_args)).float()
		)
		self.int_to_feat_matrix.requires_grad = False


class RBFFeaturizer(IntFeaturizer):
	"""
	A featurizer that puts radial basis functions evenly between 0 and max_count-1. These will have a width of
	(max_count-1) / (num_funcs) to decay to about 0.6 of its original height at reaching the next func.

	"""

	def __init__(self, num_funcs=32, max_count_int=DEFAULT_MAX_COUNT_INT, num_extra_embeddings=DEFAULT_NUM_EXTRA_EMBEDDINGS):
		"""
		:param num_funcs: number of radial basis functions to use: their width will automatically be chosen -- see class
							docstring.
		"""
		super().__init__(
			embedding_dim=num_funcs,
			max_count_int=max_count_int,
			num_extra_embeddings=num_extra_embeddings
		)
		width = (self.max_count_int - 1) / num_funcs
		centers = torch.linspace(0, self.max_count_int- 1, num_funcs)

		pre_exponential_terms = (
			-0.5
			* ((torch.arange(self.max_count_int)[:, None] - centers[None, :]) / width)
			** 2
		)
		# ^ shape: MAX_COUNT_INT x num_funcs
		feats = torch.exp(pre_exponential_terms)

		self.int_to_feat_matrix = nn.Parameter(feats.float())
		self.int_to_feat_matrix.requires_grad = False


class OneHotFeaturizer(IntFeaturizer):
	"""
	A featurizer that turns integers into their one hot encoding.

	Represents:
	 - 0 as 1000000000...
	 - 1 as 0100000000...
	 - 2 as 0010000000...
	 and so on.
	"""

	def __init__(self, max_count_int=DEFAULT_MAX_COUNT_INT, num_extra_embeddings=DEFAULT_NUM_EXTRA_EMBEDDINGS):
		super().__init__(
			embedding_dim=max_count_int,
			max_count_int=max_count_int,
			num_extra_embeddings=num_extra_embeddings
		)
		feats = torch.eye(self.max_count_int)
		self.int_to_feat_matrix = nn.Parameter(feats.float())
		self.int_to_feat_matrix.requires_grad = False


class LearnedFeaturizer(IntFeaturizer):
	"""
	Learns the features for the different integers.

	Pretty much `nn.Embedding` but we get to use the forward of the superclass which behaves a bit differently.
	"""

	def __init__(self, feature_dim=32, max_count_int=DEFAULT_MAX_COUNT_INT, num_extra_embeddings=DEFAULT_NUM_EXTRA_EMBEDDINGS):
		super().__init__(
			embedding_dim=feature_dim,
			max_count_int=max_count_int,
			num_extra_embeddings=num_extra_embeddings
		)
		weights = torch.zeros(self.max_count_int, feature_dim)
		self.int_to_feat_matrix = nn.Parameter(weights, requires_grad=True)
		nn.init.normal_(self.int_to_feat_matrix, 0.0, 1.0)


# class FloatFeaturizer(IntFeaturizer):
#	 """
#	 Norms the features
#	 """

#	 def __init__(self):
#		 # Norm vec
#		 # Placeholder..
#		 super().__init__(embedding_dim=1)
#		 self.norm_vec = torch.from_numpy(common.NORM_VEC).float()
#		 self.norm_vec = nn.Parameter(self.norm_vec)
#		 self.norm_vec.requires_grad = False

#	 def forward(self, tensor):
#		 """
#		 Convert the integer `tensor` into its new representation -- note that it gets stacked along final dimension.
#		 """
#		 tens_shape = tensor.shape
#		 out_shape = [1] * (len(tens_shape) - 1) + [-1]
#		 return tensor / self.norm_vec.reshape(*out_shape)

#	 @property
#	 def num_dim(self):
#		 return 1


def get_embedder(embedder,**kwargs):
	if embedder == "fourier":
		embedder = FourierFeaturizer(**kwargs)
	elif embedder == "rbf":
		embedder = RBFFeaturizer(**kwargs)
	elif embedder == "one-hot":
		embedder = OneHotFeaturizer(**kwargs)
	elif embedder == "learnt":
		embedder = LearnedFeaturizer(**kwargs)
	elif embedder == "float":
		raise NotImplementedError
		# embedder = FloatFeaturizer()
	elif embedder == "fourier-sines":
		embedder = FourierFeaturizerSines(**kwargs)
	elif embedder == "abs-sines":
		embedder = FourierFeaturizerAbsoluteSines(**kwargs)
	else:
		raise NotImplementedError
	return embedder



from collections import deque, defaultdict
from collections.abc import Generator
from dataclasses import dataclass
from random import choice, shuffle

import chemparse
import numpy as np
import torch
from massspecgym.models.base import Stage
from massspecgym.models.de_novo.base import DeNovoMassSpecGymModel
from rdkit import Chem
from rdkit.Chem.MolStandardize import rdMolStandardize
from rdkit.Chem.rdMolDescriptors import CalcMolFormula
from rdkit.Chem.Descriptors import ExactMolWt
from rdkit.Chem.rdchem import Mol, BondType
from copy import deepcopy
from collections import Counter
import bisect
from itertools import combinations

# type aliases for code readability
chem_element = str
number_of_atoms = int


@dataclass(frozen=True, order=True)
class ValenceAndCharge:
    """
    A data class to store valence value with the corresponding charge
    """

    valence: int
    charge: int


@dataclass(frozen=True, order=True)
class AtomWithValence:
    """
    A data class to store atom info including the computed valence
    """

    atom_type: chem_element
    atom_valence_and_charge: ValenceAndCharge


@dataclass(frozen=True, order=True)
class BondToNeighbouringAtom:
    """
    A data class to store info about the adjacent atom
    """

    adjacent_atom: AtomWithValence
    bond_type: int


@dataclass
class AtomNodeForRandomTraversal:
    """
    A data class to store atom info including the computed valence
    """

    atom_with_valence: AtomWithValence
    _remaining_node_degree: int = None
    _remaining_node_charge: int = None

    def __post_init__(self):
        """Setting up remaining node degree and charge for random traversal"""
        self._remaining_node_degree = (
            self.atom_with_valence.atom_valence_and_charge.valence
        )
        self._remaining_node_charge = (
            self.atom_with_valence.atom_valence_and_charge.charge
        )

    @property
    def remaining_node_degree(self):
        """remaining_node_degree variable getter"""
        return self._remaining_node_degree

    @remaining_node_degree.setter
    def remaining_node_degree(self, value: int):
        """remaining_node_degree variable setter"""
        self._remaining_node_degree = value

    @property
    def remaining_node_charge(self):
        """remaining_node_charge variable getter"""
        return self._remaining_node_charge

    @remaining_node_charge.setter
    def remaining_node_charge(self, value: int):
        """remaining_node_charge variable setter"""
        self._remaining_node_charge = value


def create_rdkit_molecule_from_edge_list(
    edge_list: list[tuple[int, int]], all_graph_nodes: list[AtomNodeForRandomTraversal]
) -> Mol:
    """
    A helper function converting a randomly generated edge list into rdkit.Chem.rdchem.Mol object
    @param edge_list: a list of edges, where each edge is specified by the index of its nodes
    @param all_graph_nodes: a list of all atomic nodes in the molecular graph
    """
    # first we traverse all randomly generated edges and compute bond types between each pair of atoms
    edge_2_bondtype = defaultdict(int)
    for edge_node_i, edge_node_j in edge_list:
        edge_2_bondtype[
            (min(edge_node_i, edge_node_j), max(edge_node_i, edge_node_j))
        ] += 1

    # helper routine to get the rdking enum bondtype
    def _get_rdkit_bondtype(bondtype: int) -> BondType:
        int_bondtype_2_enum = {
            1: BondType.SINGLE,
            2: BondType.DOUBLE,
            3: BondType.TRIPLE,
            4: BondType.QUADRUPLE,
            5: BondType.QUINTUPLE,
            6: BondType.HEXTUPLE,
        }
        try:
            return int_bondtype_2_enum[bondtype]
        except KeyError:
            raise NotImplementedError(f"Bond type {bondtype} is not supported")

    edge_list_rdkit = [
        (node_i, node_j, _get_rdkit_bondtype(bondtype))
        for (node_i, node_j), bondtype in edge_2_bondtype.items()
    ]
    # creating an empty editable molecule
    mol = Chem.RWMol()
    # adding the atoms to the molecule object

    # as some all_graph nodes can represent charges, we have to remember mapping of molecular atom index to
    # the corresponding atom index in all_graph_nodes
    all_graph_atom_idx_2_mol_atom_idx = {}
    for all_graph_atom_idx, atom in enumerate(all_graph_nodes):
        # ignoring charge-related graph nodes
        if atom.atom_with_valence.atom_type not in {"+", "-"}:
            all_graph_atom_idx_2_mol_atom_idx[all_graph_atom_idx] = mol.GetNumAtoms()
            next_atom = Chem.Atom(atom.atom_with_valence.atom_type)
            next_atom.SetFormalCharge(
                atom.atom_with_valence.atom_valence_and_charge.charge
            )
            mol.AddAtom(next_atom)

    # adding bonds
    for (edge_node_i, edge_node_j, bond_type) in edge_list_rdkit:
        # checking if the edge represents a charge of connected atom
        the_edge_represents_charge = len(
            {
                all_graph_nodes[node_i].atom_with_valence.atom_type
                for node_i in [edge_node_i, edge_node_j]
            }.intersection({"+", "-"})
        )
        if the_edge_represents_charge:
            # setting a charge to the corresponding atom
            for node_i in [edge_node_i, edge_node_j]:
                if all_graph_nodes[node_i].atom_with_valence.atom_type in {"+", "-"}:
                    charge_value = (
                        1
                        if all_graph_nodes[node_i].atom_with_valence.atom_type == "+"
                        else -1
                    )
                else:
                    atom_node_i = node_i
            mol.GetAtomWithIdx(
                all_graph_atom_idx_2_mol_atom_idx[atom_node_i]
            ).SetFormalCharge(charge_value)
        else:
            mol.AddBond(
                all_graph_atom_idx_2_mol_atom_idx[edge_node_i],
                all_graph_atom_idx_2_mol_atom_idx[edge_node_j],
                bond_type,
            )
    # returning the rdkit.Chem.rdchem.Mol object
    return mol.GetMol()


class RandomDeNovo(DeNovoMassSpecGymModel):
    def __init__(
        self,
        formula_known: bool = True,
        count_of_valid_valence_assignments: int = 10,
        estimate_chem_element_stats: bool = False,
        max_top_k: int = 10,
        enforce_connectivity: bool = True,
        cache_results: bool = True,
        **kwargs
    ):
        """

        @param formula_known: a boolean flag about the information available prior to generation
                              If formula_known is True, we should generate molecules with the specified formula
                              If formula_known is False, we should generate any molecule with the specified mass
        @param count_of_valid_valence_assignments: an integer controlling process of selecting valence assignment
                                                   to each atom in the generated molecule.
                                                   `count_of_valid_valence_assignments` of assignment corresponding to
                                                    the formula are generated, then one assignment is is picked at random.
                                                    The default is set to 3 for the computational speed purposes.
                                                    When setting to 1, the first feasible valence assignment will be used.
        @param estimate_chem_element_stats: a boolean flag controlling if prior information about elements' valences
                                            and bond type distributions is estimated from training data
        @param max_top_k: a maximum number of candidates to generate. If the count of valid valence assignments do
                          not allow generation of max_top_k, then less candidates are returned
        @param enforce_connectivity: a boolean flag controlling connectivity of randomly generated molecules.
                                     When it is set to True, first a random spanning tree is sampled
        @param cache_results: a boolean flag controlling caching of already generated structures.
                              When set to True, for each unique formula the set of random molecules is cached to avoid
                              recomputation.
        """
        super(RandomDeNovo, self).__init__(**kwargs)
        self.formula_known = formula_known
        self.count_of_valid_valence_assignments = count_of_valid_valence_assignments
        self.estimate_chem_element_stats = estimate_chem_element_stats
        self.max_top_k = min(max(self.top_ks), max_top_k)
        self.enforce_connectivity = enforce_connectivity
        # prior chemical knownledge about element valences
        self.element_2_valences = ELEMENT_VALENCES
        # a dictionary structure to record molecular weights with corresponding formulas from training data
        # during training steps, for each molecular weight we record all encountered formulas
        # then on training end we compute proportions of the formulas and record it as a mapping
        # mol_weight -> [[formula_1, formula_2], [proportion_of_formula_1, proportion_of_formula_2]]
        self.mol_weight_2_formulas = defaultdict(list)
        # a helper array to store sorted list of train molecular weights.
        # It will be used for the O(logn) lookup of the closest mol weight
        self.mol_weight_trn_values: list[float] = None
        # a dictionary structure for statistics about bond type distributions
        # the dictionary has the following mapping:
        # chem_element ->
        #   ValenceAndCharge ->
        #     number of already bonded atoms ->
        #       [already created BondToNeighbouringAtom] ->
        #                                           AtomWithValence ->
        #                                                          list of (bond_type, count) + total_count
        self.element_2_bond_stats = None
        # a cache with already precomputed sets of randomly generated molecules for the given formula
        self.formula_2_random_smiles = {}
        self.cache_results = cache_results

    def generator_for_splits_of_chem_element_atoms_by_possible_valences(
        self,
        atom_type: chem_element,
        possible_valences: list[ValenceAndCharge],
        atom_count: int,
        already_assigned_groups_of_atoms: dict[AtomWithValence, number_of_atoms],
    ) -> Generator[dict[AtomWithValence, number_of_atoms]]:
        """
        A recursive generator function to iterate over all possible partitions of element atoms
        into groups with different valid valences.
        Each allowed valence value can have any number from atoms, from zero up to total `atom_count`
        @param atom_type: chemical element
        @param possible_valences: a list of allowed valences
        @param atom_count: a total number of element atoms to split into valence groups
        @param already_assigned_groups_of_atoms: partial results to pass into the subsequent recursive calls

        @return A generator for lazy enumeration over all possible splits of `atom_count` atoms into subgroups
                of valid valences specified in `possible valences` parameters.
                Each return value is a dictionary, mapping atom with fixed valence to a total count of such instances
                in the molecule.

        @note In the future the method can be made into a function in a separate utils module,
        for the simplicity of codebase organization and testing purposes it's kept as the method for now
        """
        # the check for a base case of the recursion
        if atom_count == 0:
            yield already_assigned_groups_of_atoms
        elif len(possible_valences):
            # taking the first valence value from the possible ones
            next_valence = possible_valences[0]
            # iterating over possible sizes for a group of atoms with `next_valence` value of the valence
            for size_of_group in range(atom_count, -1, -1):
                # recording the assigned size of the group
                already_assigned_groups_of_atoms_next = (
                    already_assigned_groups_of_atoms.copy()
                )
                atom_with_valence = AtomWithValence(
                    atom_type=atom_type, atom_valence_and_charge=next_valence
                )
                already_assigned_groups_of_atoms_next[atom_with_valence] = size_of_group
                yield from self.generator_for_splits_of_chem_element_atoms_by_possible_valences(
                    atom_type=atom_type,
                    possible_valences=possible_valences[1:],
                    atom_count=atom_count - size_of_group,
                    already_assigned_groups_of_atoms=already_assigned_groups_of_atoms_next,
                )

    def assigner_of_valences_to_all_atoms(
        self,
        unassigned_molecule_elements_with_counts: dict[chem_element, number_of_atoms],
        already_assigned_atoms_with_valences: dict[AtomWithValence, number_of_atoms],
        common_valences_only: bool = True,
    ) -> Generator[dict[AtomWithValence, number_of_atoms]]:
        """
        A recursive function to iterate over all possible valid assignments of valences for each atom in the molecule
        @param unassigned_molecule_elements_with_counts: a dictionary representation of a molecule,
                                                         mapping each present element to a corresponding number of atoms.
                                                         The function is recursive, in the subsequence calls
                                                         the dictionary represents an yet-unprocessed submolecule
        @param already_assigned_atoms_with_valences: partial results to pass into the subsequent recursive calls,
                                                     stored as a dictionary, mapping atom with fixed valence
                                                     to a total count of such atoms in the molecule
        @param common_valences_only: a flag for using the common valence values for each element

        @return A generator for lazy enumeration over all possible assignments of all molecule atoms into subgroups
                defined by valences. Valence values are the valid ones for the corresponding chemical element.
                Each return value is a dictionary, mapping atom of specified chemical element with a fixed valence
                to a total count of such atoms in the molecule.

        @note In the future the method can be made into a function in a separate utils module,
        for the simplicity of codebase organization and testing purposes it's kept as the method for now
        """
        # the check for a base case of the recursion
        if len(unassigned_molecule_elements_with_counts) == 0:
            yield already_assigned_atoms_with_valences
        else:
            # processing the next chemical element in the molecule
            chem_element_type, atom_count = list(
                unassigned_molecule_elements_with_counts.items()
            )[0]
            # for the subsequence recursive calls the picked atom will be removed from the yet-to-be-processed
            remaining_unassigned_atoms_with_counts = (
                unassigned_molecule_elements_with_counts.copy()
            )
            del remaining_unassigned_atoms_with_counts[chem_element_type]
            # generating splits of the element count into groups with possible valences
            valences_common, valences_others = self.element_2_valences[
                chem_element_type.capitalize()
            ]
            possible_element_valences = (
                valences_common
                if common_valences_only
                else valences_common + valences_others
            )
            # we ignore "the direction" of ionic bonds, therefore we work with absolute values of valences
            possible_element_valences = map(
                lambda x: ValenceAndCharge(valence=np.abs(x.valence), charge=x.charge),
                possible_element_valences,
            )
            # we require a connected molecule graph, so we ignore possible 0 values of valences
            possible_element_valences = list(
                set(filter(lambda x: x.valence > 0, possible_element_valences))
            )
            # creating a generator for lazy enumeration over all possible splits of element atoms
            # into subgroups of possible valid valences
            valence_split_generator = (
                self.generator_for_splits_of_chem_element_atoms_by_possible_valences(
                    atom_type=chem_element_type,
                    possible_valences=possible_element_valences,
                    atom_count=atom_count,
                    already_assigned_groups_of_atoms=dict(),
                )
            )
            # iterating over splits of the element count into groups with possible valences
            for element_atoms_with_valence_2_count in valence_split_generator:
                already_assigned_atoms_with_valences_new = (
                    already_assigned_atoms_with_valences.copy()
                )
                already_assigned_atoms_with_valences_new.update(
                    element_atoms_with_valence_2_count
                )
                yield from self.assigner_of_valences_to_all_atoms(
                    unassigned_molecule_elements_with_counts=remaining_unassigned_atoms_with_counts,
                    already_assigned_atoms_with_valences=already_assigned_atoms_with_valences_new,
                    common_valences_only=common_valences_only,
                )

    def is_valence_assignment_feasible(
        self, valence_assignment: dict[AtomWithValence, number_of_atoms]
    ) -> bool:
        """
        A function for checking if the valence assignment to all molecule atoms can be feasible

        @param valence_assignment: an assignment of all molecule atoms into subgroups of plausible valences

        @note In the future the method can be made into a function in a separate utils module,
        for the simplicity of codebase organization and testing purposes it's kept as the method for now
        """
        # considering a molecule as a graph with atom being nodes and chemical bonds being edges
        # computing sum of all node degrees
        sum_of_all_node_degrees = sum(
            [
                atom.atom_valence_and_charge.valence * count_of_atoms
                for atom, count_of_atoms in valence_assignment.items()
            ]
        )
        if sum_of_all_node_degrees % 2 == 1:
            # the valence assignment is infeasible as in the graph the number of edges is half of the total degrees sum
            # therefore the sum_of_all_node_degrees must be an even number
            return False
        total_number_of_bonds = sum_of_all_node_degrees / 2
        # the total number of all atoms in the whole molecule
        total_number_of_atoms_in_molecule = sum(valence_assignment.values())
        if total_number_of_bonds < total_number_of_atoms_in_molecule - 1:
            # the valence assignment is infeasible as the molecule graph cannot be connected
            return False
        # check that charges add up to zero
        total_charge = 0
        for atom, count_of_atoms in valence_assignment.items():
            # we do not take virtual nodes for the charged molecules, we force the remaining submolecule to be neutral
            if atom.atom_type not in {"+", "-"}:
                total_charge += atom.atom_valence_and_charge.charge * count_of_atoms
        if total_charge != 0:
            return False
        return True

    def get_feasible_atom_valence_assignments(
        self, chemical_formula: str
    ) -> list[dict[AtomWithValence, number_of_atoms]]:
        """
        A function generating candidate assignments of valences to individual atoms in the molecule.
        Candidates are returned in a random order.
        @param chemical_formula: a string containing the chemical formula of the molecule

        @note In the future the method can be made into a function in a separate utils module,
        for the simplicity of codebase organization and testing purposes it's kept as the method for now
        """
        # parsing chemical formula into a dictionary of elements with corresponding counts
        element_2_count = {
            element: int(count)
            for element, count in chemparse.parse_formula(chemical_formula).items()
        }
        # checking that all input elements are valid
        for element in element_2_count.keys():
            if element.capitalize() not in self.element_2_valences:
                raise ValueError(
                    f"Found an unknown element {element.capitalize()} in the formula {chemical_formula}"
                )

        # estimate the total number of all atoms in the whole molecule
        # it will be used to check validity of the valence assignments
        total_number_of_atoms_in_molecule = sum(element_2_count.values())
        generated_candidate_valence_assignments = []
        valence_assignment_generator = self.assigner_of_valences_to_all_atoms(
            unassigned_molecule_elements_with_counts=element_2_count,
            already_assigned_atoms_with_valences=dict(),
            common_valences_only=True,
        )
        termination_assignment_value = {AtomWithValence("No more assignments", -1): -1}
        next_valence_assignment = next(
            valence_assignment_generator, termination_assignment_value
        )
        while (
            len(generated_candidate_valence_assignments)
            < self.count_of_valid_valence_assignments
            and next_valence_assignment != termination_assignment_value
        ):
            if self.is_valence_assignment_feasible(next_valence_assignment):
                generated_candidate_valence_assignments.append(next_valence_assignment)
            next_valence_assignment = next(
                valence_assignment_generator, termination_assignment_value
            )
        # if no valence assignment was found with common valences,
        # then try generating assignments including not-common valences
        if len(generated_candidate_valence_assignments) == 0:
            valence_assignment_generator = self.assigner_of_valences_to_all_atoms(
                unassigned_molecule_elements_with_counts=element_2_count,
                already_assigned_atoms_with_valences=dict(),
                common_valences_only=False,
            )
            next_valence_assignment = next(
                valence_assignment_generator, termination_assignment_value
            )
            while (
                len(generated_candidate_valence_assignments)
                < self.count_of_valid_valence_assignments
                and next_valence_assignment != termination_assignment_value
            ):
                if self.is_valence_assignment_feasible(next_valence_assignment):
                    generated_candidate_valence_assignments.append(
                        next_valence_assignment
                    )
                next_valence_assignment = next(
                    valence_assignment_generator, termination_assignment_value
                )

        if len(generated_candidate_valence_assignments) == 0:
            raise ValueError(
                f"No valence assignments can be generated for the formula {chemical_formula}"
            )
        shuffle(generated_candidate_valence_assignments)
        return generated_candidate_valence_assignments

    def sample_second_edgenode_at_random(
        self,
        edge_start_node_i: int,
        all_graph_nodes: list[AtomNodeForRandomTraversal],
        open_nodes_for_sampling: dict[str, set[int]],
        possible_candidates_type: str,
        closed_set: set[int],
        use_chem_element_stats: bool = False,
        already_connected_neighbours: list[BondToNeighbouringAtom] = None,
    ):
        """
        A function randomly sampling the second node for an edge
        @param edge_start_node_i: index of the first edge node
        @param all_graph_nodes: a list of all nodes in the molecule graph
        @param open_nodes_for_sampling: dictionary with sets of node indices which
                                                 can be considered for closing the edge.
                                                 Each set is specified by the dictionary key:
                                                     "coordinate_bond_negatively_charged_targets",
                                                     "coordinate_bond_positively_charged_targets",
                                                     "covalent_bond_targets"
        @param possible_candidates_type: the `open_nodes_for_sampling` dictionary key
        @param closed_set: closed set for traversal
        @param use_chem_element_stats: a boolean flag setting up usage of per chem. elements statistics about its bonds
        @param already_connected_neighbours: an adjacency list of already sampled neighbours
        """
        if not use_chem_element_stats:
            edge_end_node_j = choice(
                [
                    candidate_node_j
                    for candidate_node_j in open_nodes_for_sampling[
                        possible_candidates_type
                    ]
                    if candidate_node_j not in closed_set
                ]
            )
            bond_degree = 1
        else:
            # checking the current state of the atom and gathering the corresponding stats
            number_of_already_sampled_neighbours = len(already_connected_neighbours)
            # note that the graph is undirected, start-end node refers to the random traversal only
            start_atom = all_graph_nodes[edge_start_node_i]
            if self.element_2_bond_stats is None:
                raise RuntimeError(
                    "To use chem. element stats, the model has to be trained first,"
                    "to record training molecular weights with corresponding formulas."
                )
            # the structure of `self.element_2_bond_stats` is
            # chem_element ->
            #   ValenceAndCharge ->
            #     number of already bonded atoms ->
            #       [already created BondToNeighbouringAtom] ->
            #                                           AtomWithValence ->
            #                                                          list of (bond_type, count)+ total_count

            # if we don't have stats  -> fall back to sampling from all candidates
            try:
                element_stats = self.element_2_bond_stats[
                    start_atom.atom_with_valence.atom_type
                ][
                    ValenceAndCharge(
                        start_atom.atom_with_valence.atom_valence_and_charge.valence,
                        start_atom.atom_with_valence.atom_valence_and_charge.charge,
                    )
                ][
                    number_of_already_sampled_neighbours
                ][
                    tuple(sorted(already_connected_neighbours))
                ]

                full_candidates_list = []
                neighb_with_stats_candidates_list = []
                neighb_with_stats_bondcounts = []
                neighb_with_stats_bondlists = []
                # iterating over open nodes of the corresponding bond type
                for candidate_node_j in open_nodes_for_sampling[
                    possible_candidates_type
                ]:
                    if candidate_node_j not in closed_set:
                        # remembering all candidates in case no statistic-based option is there
                        full_candidates_list.append(candidate_node_j)
                        # checking if the candidate is present in element-specific bond stats
                        candidate_neighb_atom = all_graph_nodes[
                            candidate_node_j
                        ].atom_with_valence
                        if candidate_neighb_atom in element_stats:
                            neighb_with_stats_candidates_list.append(candidate_node_j)
                            bondslist, total_bond_count = element_stats[
                                candidate_neighb_atom
                            ]
                            neighb_with_stats_bondcounts.append(total_bond_count)
                            neighb_with_stats_bondlists.append(bondslist)
                # when no stats-based neighbour remain (e.g. hydrogens are not recorded in the stats)
                if len(neighb_with_stats_candidates_list) == 0:
                    edge_end_node_j = choice(full_candidates_list)
                    bond_degree = 1
                else:
                    # sampling based on frequences in bond stats
                    total_bondcount_sum = sum(neighb_with_stats_bondcounts)
                    proportions = [
                        val / total_bondcount_sum
                        for val in neighb_with_stats_bondcounts
                    ]
                    edge_end_node_j = np.random.choice(
                        neighb_with_stats_candidates_list, p=proportions
                    )
                    # getting i of the sampled neighbour to access its bond-stats
                    neighb_i = neighb_with_stats_candidates_list.index(edge_end_node_j)
                    # for the sampled end node, we sample the type of the bond based on the stats
                    bondtypes_possible = []
                    counts_of_possible_bondtypes = []
                    total_possible_bondtype_count = 0
                    # we leave only the bonds which current state of random generation allows
                    # i.e., we cannot sample a bond violating the current remaining degree of `edge_start_node_i`
                    start_node_remaining_degree = all_graph_nodes[
                        edge_start_node_i
                    ].remaining_node_degree
                    for bondtype, count in neighb_with_stats_bondlists[neighb_i]:
                        if bondtype <= start_node_remaining_degree:
                            bondtypes_possible.append(bondtype)
                            counts_of_possible_bondtypes.append(count)
                            total_possible_bondtype_count += count
                    # if no bonds can be closed for the sampled element, fall back to sampling from full candidates list
                    if len(bondtypes_possible) == 0:
                        edge_end_node_j = choice(full_candidates_list)
                        bond_degree = 1
                    else:
                        bond_degree_proportions = [
                            num / total_possible_bondtype_count
                            for num in counts_of_possible_bondtypes
                        ]
                        bond_degree = np.random.choice(
                            bondtypes_possible, p=bond_degree_proportions
                        )
                        already_connected_neighbours.append(
                            BondToNeighbouringAtom(
                                adjacent_atom=all_graph_nodes[
                                    edge_end_node_j
                                ].atom_with_valence,
                                bond_type=bond_degree,
                            )
                        )
            except:
                edge_end_node_j = choice(
                    [
                        candidate_node_j
                        for candidate_node_j in open_nodes_for_sampling[
                            possible_candidates_type
                        ]
                        if candidate_node_j not in closed_set
                    ]
                )
                bond_degree = 1
        return edge_end_node_j, bond_degree, already_connected_neighbours

    def sample_edge_at_random(
        self,
        all_graph_nodes: list[AtomNodeForRandomTraversal],
        open_nodes_for_sampling: dict[str, set[int]],
        edge_start_node_i: int = None,
        closed_set: set[int] = None,
        use_chem_element_stats: bool = False,
        atom_2_already_connected_neighbours: list[list[BondToNeighbouringAtom]] = None,
    ) -> tuple[tuple[int, int], list[AtomNodeForRandomTraversal], set[int]]:
        """
        Helper function to filter atoms suitable for generation of a random bond with `edge_start_node_i`
        and sampling a random edge
        @param all_graph_nodes: a list of all nodes in the molecule graph
        @param edge_start_node_i: index of the first edge node
        @param open_nodes_for_sampling: dictionary with sets of node indices which
                                                 can be considered for closing the edge.
                                                 Each set is specified by the dictionary key:
                                                     "coordinate_bond_negatively_charged_targets",
                                                     "coordinate_bond_positively_charged_targets",
                                                     "covalent_bond_targets"
        @param use_chem_element_stats: a boolean flag setting up usage of per chem. elements statistics about its bonds
        @param closed_set: closed set for traversal
        @param atom_2_already_connected_neighbours: a mapping from atom to its adjacency list of already sampled neighbours
        @return: a sampled edge and updated structures `all_graph_nodes`, `open_nodes_for_sampling`
        """
        # sample the start node for the edge if it's not specified
        if edge_start_node_i is None:
            edge_start_node_i = choice(
                sum(map(list, open_nodes_for_sampling.values()), [])
            )
        if closed_set is None:
            closed_set = {edge_start_node_i}
        # check if the start edge atom has the charge and therefore can form coordinate bond
        can_form_coordinate_bond = (
            all_graph_nodes[edge_start_node_i].remaining_node_charge != 0
        )
        # if possible, create coordinate bond at random
        is_bond_coordinate = can_form_coordinate_bond and np.random.rand() < 0.5
        if is_bond_coordinate:
            start_node_charge_sign = np.sign(
                all_graph_nodes[edge_start_node_i].remaining_node_charge
            )
            # if for the coordinate bond one atom is positively charged, then another must be charged negatively
            if start_node_charge_sign > 0:
                possible_candidates_type = "coordinate_bond_neg_charged_targets"
            else:
                possible_candidates_type = "coordinate_bond_pos_charged_targets"
        else:
            possible_candidates_type = "covalent_bond_targets"

        (
            edge_end_node_j,
            node_degree_reduction,
            atom_2_already_connected_neighbours[edge_start_node_i],
        ) = self.sample_second_edgenode_at_random(
            edge_start_node_i,
            all_graph_nodes,
            open_nodes_for_sampling,
            possible_candidates_type,
            closed_set,
            use_chem_element_stats,
            atom_2_already_connected_neighbours[edge_start_node_i],
        )

        # decrease the node degrees correspondingly
        for node_of_a_new_edge_i in [edge_start_node_i, edge_end_node_j]:
            all_graph_nodes[
                node_of_a_new_edge_i
            ].remaining_node_degree -= node_degree_reduction
            # if all bonds are created for the particular atom, it is no more open for traversal
            if all_graph_nodes[node_of_a_new_edge_i].remaining_node_degree <= 0:
                for candidates_type in open_nodes_for_sampling.keys():
                    if node_of_a_new_edge_i in open_nodes_for_sampling[candidates_type]:
                        open_nodes_for_sampling[candidates_type].remove(
                            node_of_a_new_edge_i
                        )
            # if the added bond was coordinate, modify the remaining charges correspondingly
            elif is_bond_coordinate:
                new_charge_abs_value = (
                    np.abs(all_graph_nodes[node_of_a_new_edge_i].remaining_node_charge)
                    - 1
                )
                # check if the node still can form coordinate bonds
                if new_charge_abs_value == 0:
                    for candidates_type in [
                        "coordinate_bond_neg_charged_targets",
                        "coordinate_bond_pos_charged_targets",
                    ]:
                        if (
                            node_of_a_new_edge_i
                            in open_nodes_for_sampling[candidates_type]
                        ):
                            open_nodes_for_sampling[candidates_type].remove(
                                node_of_a_new_edge_i
                            )
                else:
                    charge_sign = np.sign(
                        all_graph_nodes[node_of_a_new_edge_i].remaining_node_charge
                    )
                    all_graph_nodes[node_of_a_new_edge_i].remaining_node_charge = (
                        charge_sign * new_charge_abs_value
                    )
        return (
            (edge_start_node_i, edge_end_node_j),
            all_graph_nodes,
            open_nodes_for_sampling,
        )

    def generate_random_molecule_graphs_via_traversal(
        self,
        chemical_formula: str,
        max_number_of_retries_per_valence_assignment: int = 100,
    ) -> list[Mol]:
        """
        A function generating random molecule graph(s).
        The generation process ensures that each graph is connected.
        If any of the `self.count_of_valid_valence_assignments` enables it,
        the function returns graph(s) without self-loops.

        @param chemical_formula: a string containing the chemical formula of the molecule
        @param max_number_of_retries_per_valence_assignment: a max count of attempts to generate a random spanning tree
                                                             for a given potentially feasible valence assignment

        @note In the future the method can be made into a function in a separate utils module,
        for the simplicity of codebase organization and testing purposes it's kept as the method for now
        """
        # check if for the input formula the random structures have been already generated
        if self.cache_results and chemical_formula in self.formula_2_random_smiles:
            return self.formula_2_random_smiles[chemical_formula]

        # get candidate partitions of all molecule atoms into valences
        candidate_valence_assignments = self.get_feasible_atom_valence_assignments(
            chemical_formula
        )
        # iterate over each valence assignment to all atoms, the order is random
        assert (
            len(candidate_valence_assignments) > 0
        ), f"No potentially feasible atom valence assignment for {chemical_formula}"
        # number of iteration over feasible valence assignments
        num_of_iterations_over_splits_into_valences = int(
            np.ceil(self.max_top_k / len(candidate_valence_assignments))
        )
        generated_molecules = []

        # we request to generate self.max_top_k molecule(s)
        while len(generated_molecules) < self.max_top_k:
            for _ in range(num_of_iterations_over_splits_into_valences):
                for valence_assignment in candidate_valence_assignments:
                    # first randomly create a spanning tree of the molecule graph, to ensure the connectivity of molecule.
                    # The feasibility check `self.is_valence_assignment_feasible` inside the
                    # `self.get_feasible_atom_valence_assignments` function should ensure the possibility to create the tree.
                    spanning_tree_was_generated = False
                    spanning_tree_generation_attempts = 0
                    while (
                        not spanning_tree_was_generated
                        and spanning_tree_generation_attempts
                        < max_number_of_retries_per_valence_assignment
                    ):
                        spanning_tree_generation_attempts += 1
                        # we optimistically set the value of `spanning_tree_was_generated` to True,
                        # If the current traversal do not lead to a spanning tree,
                        # then `spanning_tree_was_generated` is set to False in the code below
                        spanning_tree_was_generated = True

                        # prepare node list for a random edges generation
                        all_graph_nodes = []
                        for (
                            atom_with_valence,
                            num_of_atoms_in_molecule,
                        ) in valence_assignment.items():
                            for _ in range(num_of_atoms_in_molecule):
                                all_graph_nodes.append(
                                    AtomNodeForRandomTraversal(
                                        atom_with_valence=atom_with_valence
                                    )
                                )

                        # a helper structure to record already sampled bonds
                        # it is used only if we use estimated chem elements stats in the generation process
                        atom_2_already_connected_neighbours = [
                            [] for _ in range(len(all_graph_nodes))
                        ]

                        # recording sets of nodes available for random sampling of covalent and coordinate bonds
                        coordinate_bond_neg_charged_targets = {
                            node_i
                            for node_i, node in enumerate(all_graph_nodes)
                            if np.sign(node.remaining_node_charge) == -1
                        }
                        coordinate_bond_pos_charged_targets = {
                            node_i
                            for node_i, node in enumerate(all_graph_nodes)
                            if np.sign(node.remaining_node_charge) == 1
                        }
                        covalent_bond_targets = {
                            node_i
                            for node_i, node in enumerate(all_graph_nodes)
                            if node.remaining_node_charge == 0
                            or node.remaining_node_degree
                            > np.abs(node.remaining_node_charge)
                        }

                        open_nodes_for_sampling = {
                            "coordinate_bond_neg_charged_targets": coordinate_bond_neg_charged_targets,
                            "coordinate_bond_pos_charged_targets": coordinate_bond_pos_charged_targets,
                            "covalent_bond_targets": covalent_bond_targets,
                        }

                        # the final edge list will be stored into the variable below.
                        # An edge is defined by a pair of position indices in the `all_graph_nodes` list
                        edge_list = []

                        # the nodes already included into the spanning tree
                        # the set is used for quick blacklisting, while the list is used for possible backtracking when
                        (
                            spanning_tree_visited_nodes_set,
                            spanning_tree_traversal_list,
                        ) = (
                            set(),
                            deque(),
                        )
                        # sample a random start of spanning tree generation
                        edge_start_node_i = choice(list(range(len(all_graph_nodes))))
                        spanning_tree_visited_nodes_set.add(edge_start_node_i)
                        spanning_tree_traversal_list.append(edge_start_node_i)
                        while self.enforce_connectivity and len(
                            spanning_tree_visited_nodes_set
                        ) < len(all_graph_nodes):
                            # check if the start edge atom has the charge and therefore can form coordinate bond
                            try:
                                (
                                    (edge_start_node_i, edge_end_node_i),
                                    all_graph_nodes,
                                    open_nodes_for_sampling,
                                ) = self.sample_edge_at_random(
                                    all_graph_nodes,
                                    open_nodes_for_sampling,
                                    edge_start_node_i=edge_start_node_i,
                                    closed_set=spanning_tree_visited_nodes_set,
                                    use_chem_element_stats=self.estimate_chem_element_stats,
                                    atom_2_already_connected_neighbours=atom_2_already_connected_neighbours,
                                )
                            except IndexError:
                                spanning_tree_was_generated = False
                                break
                            # note that the graph is undirected, start-end node refers to the random traversal only
                            edge_list.append((edge_start_node_i, edge_end_node_i))
                            # recording the node added to the random spanning tree
                            spanning_tree_visited_nodes_set.add(edge_end_node_i)
                            spanning_tree_traversal_list.append(edge_end_node_i)

                            # finding a start node for the next sampled edge.
                            # We have to ensure that such a node still has some degree not covered by sampling nodes.
                            # For that, we might need to backtrack.
                            candidate_for_start_node_i = edge_end_node_i
                            try:
                                while (
                                    all_graph_nodes[
                                        candidate_for_start_node_i
                                    ].remaining_node_degree
                                    == 0
                                ):
                                    spanning_tree_traversal_list.pop()
                                    candidate_for_start_node_i = (
                                        spanning_tree_traversal_list[-1]
                                    )
                            except IndexError:
                                spanning_tree_was_generated = False
                                break
                            edge_start_node_i = candidate_for_start_node_i

                    # after the spanning tree edges were sampled,
                    # now we randomly connect nodes with remaining degrees yet uncovered by sampled bonds
                    while sum(map(len, open_nodes_for_sampling.values())) >= 2:
                        try:
                            (
                                (edge_start_node_i, edge_end_node_i),
                                all_graph_nodes,
                                open_nodes_for_sampling,
                            ) = self.sample_edge_at_random(
                                all_graph_nodes,
                                open_nodes_for_sampling,
                                use_chem_element_stats=self.estimate_chem_element_stats,
                                atom_2_already_connected_neighbours=atom_2_already_connected_neighbours,
                            )
                        except IndexError:
                            break
                        edge_list.append((edge_start_node_i, edge_end_node_i))

                    # if all nodes were covered by edges without self-loops, then we remember the generated molecule
                    if sum(map(len, open_nodes_for_sampling.values())) == 0:
                        generated_molecules.append(
                            create_rdkit_molecule_from_edge_list(
                                edge_list, all_graph_nodes
                            )
                        )
                        if len(generated_molecules) == self.max_top_k:
                            if self.cache_results:
                                self.formula_2_random_smiles[
                                    chemical_formula
                                ] = generated_molecules
                            return generated_molecules
        if self.cache_results:
            self.formula_2_random_smiles[chemical_formula] = generated_molecules
        return generated_molecules

    def training_step(
        self, batch: dict, batch_idx: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        # recording statistics about chemical elements
        if self.estimate_chem_element_stats:
            if self.element_2_bond_stats is None:
                self.element_2_bond_stats = defaultdict(dict)
            for mol_smiles in batch["mol"]:
                molecule = Chem.MolFromSmiles(mol_smiles)
                # in order to work with double and single bonds instead of aromatic
                Chem.Kekulize(molecule, clearAromaticFlags=True)
                # we add hydrogen atoms (which are ommited by default)
                molecule = Chem.AddHs(molecule)
                formula = CalcMolFormula(molecule)
                for atom in molecule.GetAtoms():
                    valence = atom.GetTotalValence()
                    charge = atom.GetFormalCharge()
                    valence_charge = ValenceAndCharge(valence, charge)
                    chem_element_type = atom.GetSymbol()
                    atom_bonds = atom.GetBonds()
                    if (
                        valence_charge
                        not in self.element_2_bond_stats[chem_element_type]
                    ):
                        # for each value of atom's valence, number of neighbours we will count types of neighbouring atoms
                        self.element_2_bond_stats[chem_element_type][
                            valence_charge
                        ] = dict()

                    all_atom_neighbours = set()
                    for bond in atom_bonds:
                        start_atom_idx = atom.GetIdx()
                        end_atom = [
                            _atom
                            for _atom in [bond.GetBeginAtom(), bond.GetEndAtom()]
                            if _atom.GetIdx() != start_atom_idx
                        ][0]
                        all_atom_neighbours.add(
                            (
                                end_atom.GetSymbol(),
                                end_atom.GetTotalValence(),
                                end_atom.GetFormalCharge(),
                                bond.GetBondTypeAsDouble(),
                            )
                        )

                    all_neighbour_subsets = [
                        [set(subset) for subset in combinations(all_atom_neighbours, r)]
                        for r in range(len(all_atom_neighbours))
                    ]
                    for neighbour_subsets_of_fixed_size in all_neighbour_subsets:
                        subset_size = len(neighbour_subsets_of_fixed_size[0])
                        # for each number of already connected atoms we record the neighbours and then possible bonds yet to be closed
                        if (
                            subset_size
                            not in self.element_2_bond_stats[chem_element_type][
                                valence_charge
                            ]
                        ):
                            self.element_2_bond_stats[chem_element_type][
                                valence_charge
                            ][subset_size] = dict()
                        for neighbours in neighbour_subsets_of_fixed_size:
                            neighbours_tuple = tuple(sorted(neighbours))
                            if (
                                neighbours_tuple
                                not in self.element_2_bond_stats[chem_element_type][
                                    valence_charge
                                ][subset_size]
                            ):
                                self.element_2_bond_stats[chem_element_type][
                                    valence_charge
                                ][subset_size][neighbours_tuple] = defaultdict(int)
                            remaining_bonds = all_atom_neighbours.difference(neighbours)
                            for remaining_bonded_neighbour in remaining_bonds:
                                self.element_2_bond_stats[chem_element_type][
                                    valence_charge
                                ][subset_size][neighbours_tuple][
                                    remaining_bonded_neighbour
                                ] += 1

        # recording molecular weight
        for mol_smiles in batch["mol"]:
            molecule = Chem.MolFromSmiles(mol_smiles)
            formula = CalcMolFormula(molecule)
            weight = ExactMolWt(molecule)
            self.mol_weight_2_formulas[weight].append(formula)
        # Random baseline, so we return a dummy loss
        loss = torch.tensor(0.0, requires_grad=True)
        return dict(loss=loss, mols_pred=["C"])

    def on_train_end(self) -> None:
        # for each molecular weight we compute proportions of recorded molecular formulas
        molecular_weight_2_formula_counts = {
            weight: Counter(formulas)
            for weight, formulas in self.mol_weight_2_formulas.items()
        }
        weight_2_formula_proportions = {}
        for weight, formula_2_count in molecular_weight_2_formula_counts.items():
            total_count = sum(formula_2_count.values())
            weight_2_formula_proportions[weight] = {
                formula: count / total_count
                for formula, count in formula_2_count.items()
            }
        # for consequent sampling using numpy.random.choice function, we store the results in the format
        # weight -> [[formula_1, formula_2], [proportion_of_formula_1, proportion_of_formula_2]]
        self.mol_weight_2_formulas = {
            weight: [
                list(formula_2_proportions.keys()),
                list(formula_2_proportions.values()),
            ]
            for weight, formula_2_proportions in weight_2_formula_proportions.items()
        }
        # storing weights in the sorted list for the logarithmic time look-up of the closest weight value
        self.mol_weight_trn_values = sorted(self.mol_weight_2_formulas.keys())

        # if chem element stats are used, then the corresponding data structure is reformated in accordance with the
        # description from docstring to the class __init__:
        # chem_element ->
        #   ValenceAndCharge ->
        #     number of already bonded atoms ->
        #       [already created BondToNeighbouringAtom] ->
        #                                           AtomWithValence ->
        #                                                          list of (bond_type, count) + total_count
        if self.estimate_chem_element_stats:
            element_2_bond_stats = defaultdict(dict)
            for (
                chem_element,
                valence_charge_2_stats,
            ) in self.element_2_bond_stats.items():
                for valence_charge, num_bonds_2_stats in valence_charge_2_stats.items():
                    element_2_bond_stats[chem_element][valence_charge] = dict()
                    for num_bonds, bonds_2_stats in num_bonds_2_stats.items():
                        element_2_bond_stats[chem_element][valence_charge][
                            num_bonds
                        ] = dict()
                        for (
                            bonds,
                            neighb_atom_with_valence_2_stats,
                        ) in bonds_2_stats.items():
                            present_bonds_sorted = tuple(
                                sorted(
                                    [
                                        BondToNeighbouringAtom(
                                            adjacent_atom=AtomWithValence(
                                                atom_type=bond[0],
                                                atom_valence_and_charge=ValenceAndCharge(
                                                    valence=bond[1], charge=bond[2]
                                                ),
                                            ),
                                            bond_type=bond[3],
                                        )
                                        for bond in bonds
                                    ]
                                )
                            )
                            element_2_bond_stats[chem_element][valence_charge][
                                num_bonds
                            ][present_bonds_sorted] = defaultdict(list)
                            for (
                                neighb_atom_type,
                                neighb_atom_valence,
                                neighb_atom_charge,
                                bondtype,
                            ), count in neighb_atom_with_valence_2_stats.items():
                                neighbouring_atom = AtomWithValence(
                                    atom_type=neighb_atom_type,
                                    atom_valence_and_charge=ValenceAndCharge(
                                        valence=neighb_atom_valence,
                                        charge=neighb_atom_charge,
                                    ),
                                )
                                element_2_bond_stats[chem_element][valence_charge][
                                    num_bonds
                                ][present_bonds_sorted][neighbouring_atom].append(
                                    (bondtype, count)
                                )
                            # computing total count of all bound per neighbouring atom
                            for (
                                neighbouring_atom,
                                list_of_bondtype_counts,
                            ) in element_2_bond_stats[chem_element][valence_charge][
                                num_bonds
                            ][
                                present_bonds_sorted
                            ].items():
                                total_count_of_bonds = sum(
                                    map(lambda x: x[1], list_of_bondtype_counts)
                                )
                                element_2_bond_stats[chem_element][valence_charge][
                                    num_bonds
                                ][present_bonds_sorted][neighbouring_atom] = (
                                    list_of_bondtype_counts,
                                    total_count_of_bonds,
                                )
            self.element_2_bond_stats = element_2_bond_stats

    def sample_formula_with_the_closest_molecular_weight(
        self, molecular_weight: float
    ) -> str:
        """
        A method sampling chemical formula observed in training data with the closest weight to `molecular_weight`
        @param molecular_weight: Molecular weight of a structure to be generated
        """
        if self.mol_weight_trn_values is None:
            raise RuntimeError(
                "For random denovo generation without known formula, the model has to be trained first,"
                "to record training molecular weights with corresponding formulas."
            )
        # finding a place in the sorted array for insertion of the `molecular_weight`, while preserving sorted order
        idx_of_closest_larger = bisect.bisect_left(
            self.mol_weight_trn_values, molecular_weight
        )
        # check if the exact same molecular weight was observed in training data, otherwise select the closest weight
        if molecular_weight == self.mol_weight_trn_values[idx_of_closest_larger]:
            idx_of_closest = idx_of_closest_larger
        elif idx_of_closest_larger > 0:
            # determining the closest molecular weight out of both neighbours
            idx_of_closest_smaller = idx_of_closest_larger - 1
            weight_difference_with_smaller_neighbour = (
                molecular_weight - self.mol_weight_trn_values[idx_of_closest_smaller]
            )
            weight_difference_with_larger_neighbour = (
                self.mol_weight_trn_values[idx_of_closest_larger] - molecular_weight
            )
            if (
                weight_difference_with_larger_neighbour
                < weight_difference_with_smaller_neighbour
            ):
                idx_of_closest = idx_of_closest_larger
            else:
                idx_of_closest = idx_of_closest_smaller
        else:
            idx_of_closest = 0
        # the value of the molecular weight observed in training labels, which is the closest to `molecular_weight`
        closest_observed_molecular_weight = self.mol_weight_trn_values[idx_of_closest]
        # getting chemical formulas observed for this molecular weight
        # self.mol_weight_2_formulas is a dictionary containing the following mapping
        #  weight -> [[formula_1, formula_2], [proportion_of_formula_1, proportion_of_formula_2]]
        feasible_formulas, formula_proportions = self.mol_weight_2_formulas[
            closest_observed_molecular_weight
        ]
        # if just one formula is known, it is returned directly
        if len(feasible_formulas) == 1:
            return feasible_formulas[0]
        # otherwise we randomly sample in accordance with proportions
        return np.random.choice(feasible_formulas, p=formula_proportions)

    def step(
        self, batch: dict, stage: Stage = Stage.NONE
    ) -> tuple[torch.Tensor, torch.Tensor]:
        mols = batch["mol"]  # List of SMILES of length batch_size

        # If formula_known is True, we should generate molecules with the same formula as label (`mols` above)
        # If formula_known is False, we should generate any molecule with the same mass as label

        # obtaining molecule objects from SMILES
        molecules = [Chem.MolFromSmiles(smiles) for smiles in mols]
        # getting the formulas
        if self.formula_known:
            formulas = [CalcMolFormula(molecule) for molecule in molecules]
        else:
            molecular_weights = [ExactMolWt(molecule) for molecule in molecules]
            formulas = [
                self.sample_formula_with_the_closest_molecular_weight(mol_weight)
                for mol_weight in molecular_weights
            ]
        # (bs, k) list of rdkit molecules
        mols_pred = [
            self.generate_random_molecule_graphs_via_traversal(formula)
            for formula in formulas
        ]

        for predicted_mol_group in mols_pred:
            for mol in predicted_mol_group:
                Chem.RemoveHs(mol)

        # list of predicted smiles
        smiles_pred = [
            [
                Chem.MolToSmiles(mol_candidate)
                for mol_candidate in candidates_per_input_mol
            ]
            for candidates_per_input_mol in mols_pred
        ]

        # Random baseline, so we return a dummy loss
        loss = torch.tensor(0.0, requires_grad=True)
        return dict(loss=loss, mols_pred=smiles_pred)

    def configure_optimizers(self):
        # No optimizer needed for a random baseline
        return None


# element valences taken from sources like https://sciencenotes.org/element-valency-pdf
# the first list contains the typical valences, each tuple is a valence value with the corresponding charge
ELEMENT_VALENCES = {
    "H": (
        [ValenceAndCharge(valence=1, charge=0)],
        [ValenceAndCharge(valence=0, charge=0), ValenceAndCharge(valence=1, charge=-1)],
    ),
    "He": ([ValenceAndCharge(valence=0, charge=0)], []),
    "Li": (
        [ValenceAndCharge(valence=1, charge=0)],
        [ValenceAndCharge(valence=1, charge=-1)],
    ),
    "Be": ([ValenceAndCharge(valence=2, charge=0)], []),
    "B": (
        [ValenceAndCharge(valence=3, charge=0), ValenceAndCharge(valence=4, charge=-1)],
        [ValenceAndCharge(valence=2, charge=0), ValenceAndCharge(valence=1, charge=0)],
    ),
    "C": (
        [ValenceAndCharge(valence=4, charge=0)],
        [
            ValenceAndCharge(valence=3, charge=-1),
            ValenceAndCharge(valence=2, charge=0),
            ValenceAndCharge(valence=2, charge=-1),
            ValenceAndCharge(valence=1, charge=0),
            ValenceAndCharge(valence=1, charge=-1),
        ],
    ),
    "N": (
        [ValenceAndCharge(valence=3, charge=0), ValenceAndCharge(valence=4, charge=1)],
        [
            ValenceAndCharge(valence=2, charge=-1),
            ValenceAndCharge(valence=5, charge=0),
            ValenceAndCharge(valence=1, charge=0),
            ValenceAndCharge(valence=0, charge=0),
            ValenceAndCharge(valence=1, charge=-1),
        ],
    ),
    "O": (
        [ValenceAndCharge(valence=2, charge=0), ValenceAndCharge(valence=1, charge=-1)],
        [ValenceAndCharge(valence=3, charge=1)],
    ),
    "F": ([ValenceAndCharge(valence=1, charge=0)], []),
    "Ne": ([ValenceAndCharge(valence=0, charge=0)], []),
    "Na": (
        [ValenceAndCharge(valence=1, charge=0)],
        [],
    ),
    "Mg": ([ValenceAndCharge(valence=2, charge=0)], []),
    "Al": (
        [ValenceAndCharge(valence=3, charge=0)],
        [ValenceAndCharge(valence=1, charge=0)],
    ),
    "Si": (
        [ValenceAndCharge(valence=4, charge=0)],
        [],
    ),
    "P": (
        [ValenceAndCharge(valence=5, charge=0)],
        [
            ValenceAndCharge(valence=4, charge=1),
            ValenceAndCharge(valence=3, charge=0),
            ValenceAndCharge(valence=2, charge=0),
            ValenceAndCharge(valence=1, charge=0),
        ],
    ),
    "S": (
        [ValenceAndCharge(valence=2, charge=0), ValenceAndCharge(valence=6, charge=0)],
        [
            ValenceAndCharge(valence=4, charge=0),
            ValenceAndCharge(valence=1, charge=-1),
            ValenceAndCharge(valence=3, charge=1),
        ],
    ),
    "Cl": (
        [ValenceAndCharge(valence=1, charge=0)],
        [],
    ),
    "Ar": ([ValenceAndCharge(valence=0, charge=0)], []),
    "K": (
        [ValenceAndCharge(valence=1, charge=0)],
        [],
    ),
    "Ca": ([ValenceAndCharge(valence=2, charge=0)], []),
    "Sc": (
        [ValenceAndCharge(valence=3, charge=0)],
        [ValenceAndCharge(valence=2, charge=0), ValenceAndCharge(valence=1, charge=0)],
    ),
    "Ti": (
        [ValenceAndCharge(valence=4, charge=0)],
        [
            ValenceAndCharge(valence=3, charge=0),
            ValenceAndCharge(valence=2, charge=0),
            ValenceAndCharge(valence=0, charge=0),
        ],
    ),
    "V": (
        [
            ValenceAndCharge(valence=5, charge=0),
            ValenceAndCharge(valence=4, charge=0),
            ValenceAndCharge(valence=3, charge=0),
        ],
        [
            ValenceAndCharge(valence=2, charge=0),
            ValenceAndCharge(valence=1, charge=0),
            ValenceAndCharge(valence=0, charge=0),
        ],
    ),
    "Cr": (
        [
            ValenceAndCharge(valence=6, charge=0),
            ValenceAndCharge(valence=3, charge=0),
            ValenceAndCharge(valence=2, charge=0),
        ],
        [
            ValenceAndCharge(valence=5, charge=0),
            ValenceAndCharge(valence=4, charge=0),
            ValenceAndCharge(valence=1, charge=0),
            ValenceAndCharge(valence=0, charge=0),
        ],
    ),
    "Mn": (
        [
            ValenceAndCharge(valence=7, charge=0),
            ValenceAndCharge(valence=4, charge=0),
            ValenceAndCharge(valence=2, charge=0),
        ],
        [
            ValenceAndCharge(valence=6, charge=0),
            ValenceAndCharge(valence=5, charge=0),
            ValenceAndCharge(valence=3, charge=0),
            ValenceAndCharge(valence=1, charge=0),
            ValenceAndCharge(valence=0, charge=0),
        ],
    ),
    "Fe": (
        [ValenceAndCharge(valence=2, charge=0), ValenceAndCharge(valence=3, charge=0)],
        [
            ValenceAndCharge(valence=6, charge=0),
            ValenceAndCharge(valence=5, charge=0),
            ValenceAndCharge(valence=4, charge=0),
            ValenceAndCharge(valence=1, charge=0),
            ValenceAndCharge(valence=0, charge=0),
        ],
    ),
    "Co": (
        [ValenceAndCharge(valence=2, charge=0), ValenceAndCharge(valence=3, charge=0)],
        [
            ValenceAndCharge(valence=5, charge=0),
            ValenceAndCharge(valence=4, charge=0),
            ValenceAndCharge(valence=1, charge=0),
            ValenceAndCharge(valence=0, charge=0),
        ],
    ),
    "Ni": (
        [ValenceAndCharge(valence=2, charge=0)],
        [
            ValenceAndCharge(valence=6, charge=0),
            ValenceAndCharge(valence=4, charge=0),
            ValenceAndCharge(valence=3, charge=0),
            ValenceAndCharge(valence=1, charge=0),
            ValenceAndCharge(valence=0, charge=0),
        ],
    ),
    "Cu": (
        [ValenceAndCharge(valence=2, charge=0), ValenceAndCharge(valence=1, charge=0)],
        [
            ValenceAndCharge(valence=4, charge=0),
            ValenceAndCharge(valence=3, charge=0),
            ValenceAndCharge(valence=0, charge=0),
        ],
    ),
    "Zn": (
        [ValenceAndCharge(valence=2, charge=0)],
        [ValenceAndCharge(valence=1, charge=0), ValenceAndCharge(valence=0, charge=0)],
    ),
    "Ga": (
        [ValenceAndCharge(valence=3, charge=0)],
        [ValenceAndCharge(valence=2, charge=0), ValenceAndCharge(valence=1, charge=0)],
    ),
    "Ge": (
        [ValenceAndCharge(valence=4, charge=0)],
        [
            ValenceAndCharge(valence=3, charge=0),
            ValenceAndCharge(valence=2, charge=0),
            ValenceAndCharge(valence=1, charge=0),
        ],
    ),
    "As": (
        [ValenceAndCharge(valence=5, charge=0), ValenceAndCharge(valence=4, charge=1)],
        [],
    ),
    "Se": (
        [ValenceAndCharge(valence=2, charge=0)],
        [],
    ),
    "Br": (
        [ValenceAndCharge(valence=1, charge=0)],
        [],
    ),
    "Kr": (
        [ValenceAndCharge(valence=0, charge=0)],
        [ValenceAndCharge(valence=2, charge=0)],
    ),
    "Rb": (
        [ValenceAndCharge(valence=1, charge=0)],
        [],
    ),
    "Sr": ([ValenceAndCharge(valence=2, charge=0)], []),
    "Y": (
        [ValenceAndCharge(valence=3, charge=0)],
        [ValenceAndCharge(valence=2, charge=0)],
    ),
    "Zr": (
        [ValenceAndCharge(valence=4, charge=0)],
        [
            ValenceAndCharge(valence=3, charge=0),
            ValenceAndCharge(valence=2, charge=0),
            ValenceAndCharge(valence=1, charge=0),
            ValenceAndCharge(valence=0, charge=0),
        ],
    ),
    "Nb": (
        [ValenceAndCharge(valence=5, charge=0)],
        [
            ValenceAndCharge(valence=4, charge=0),
            ValenceAndCharge(valence=3, charge=0),
            ValenceAndCharge(valence=2, charge=0),
            ValenceAndCharge(valence=1, charge=0),
            ValenceAndCharge(valence=0, charge=0),
        ],
    ),
    "Mo": (
        [ValenceAndCharge(valence=6, charge=0), ValenceAndCharge(valence=4, charge=0)],
        [
            ValenceAndCharge(valence=5, charge=0),
            ValenceAndCharge(valence=3, charge=0),
            ValenceAndCharge(valence=2, charge=0),
            ValenceAndCharge(valence=1, charge=0),
            ValenceAndCharge(valence=0, charge=0),
        ],
    ),
    "Tc": (
        [ValenceAndCharge(valence=7, charge=0), ValenceAndCharge(valence=4, charge=0)],
        [
            ValenceAndCharge(valence=6, charge=0),
            ValenceAndCharge(valence=5, charge=0),
            ValenceAndCharge(valence=3, charge=0),
            ValenceAndCharge(valence=2, charge=0),
            ValenceAndCharge(valence=1, charge=0),
            ValenceAndCharge(valence=0, charge=0),
        ],
    ),
    "Ru": (
        [ValenceAndCharge(valence=4, charge=0), ValenceAndCharge(valence=3, charge=0)],
        [
            ValenceAndCharge(valence=8, charge=0),
            ValenceAndCharge(valence=7, charge=0),
            ValenceAndCharge(valence=6, charge=0),
            ValenceAndCharge(valence=5, charge=0),
            ValenceAndCharge(valence=2, charge=0),
            ValenceAndCharge(valence=1, charge=0),
            ValenceAndCharge(valence=0, charge=0),
        ],
    ),
    "Rh": (
        [ValenceAndCharge(valence=3, charge=0)],
        [
            ValenceAndCharge(valence=6, charge=0),
            ValenceAndCharge(valence=5, charge=0),
            ValenceAndCharge(valence=4, charge=0),
            ValenceAndCharge(valence=2, charge=0),
            ValenceAndCharge(valence=1, charge=0),
            ValenceAndCharge(valence=0, charge=0),
        ],
    ),
    "Pd": (
        [ValenceAndCharge(valence=4, charge=0), ValenceAndCharge(valence=2, charge=0)],
        [ValenceAndCharge(valence=0, charge=0)],
    ),
    "Ag": (
        [ValenceAndCharge(valence=1, charge=0)],
        [
            ValenceAndCharge(valence=3, charge=0),
            ValenceAndCharge(valence=2, charge=0),
            ValenceAndCharge(valence=0, charge=0),
        ],
    ),
    "Cd": (
        [ValenceAndCharge(valence=2, charge=0)],
        [ValenceAndCharge(valence=1, charge=0)],
    ),
    "In": (
        [ValenceAndCharge(valence=3, charge=0)],
        [ValenceAndCharge(valence=2, charge=0), ValenceAndCharge(valence=1, charge=0)],
    ),
    "Sn": (
        [ValenceAndCharge(valence=2, charge=0)],
        [ValenceAndCharge(valence=4, charge=0)],
    ),
    "Sb": (
        [ValenceAndCharge(valence=3, charge=0)],
        [ValenceAndCharge(valence=5, charge=0), ValenceAndCharge(valence=3, charge=-1)],
    ),
    "Te": (
        [ValenceAndCharge(valence=4, charge=0)],
        [ValenceAndCharge(valence=2, charge=0), ValenceAndCharge(valence=6, charge=0)],
    ),
    "I": (
        [ValenceAndCharge(valence=1, charge=0)],
        [
            ValenceAndCharge(valence=1, charge=0),
            ValenceAndCharge(valence=3, charge=0),
            ValenceAndCharge(valence=7, charge=0),
            ValenceAndCharge(valence=0, charge=0),
        ],
    ),
    "Xe": (
        [ValenceAndCharge(valence=0, charge=0)],
        [
            ValenceAndCharge(valence=2, charge=0),
            ValenceAndCharge(valence=4, charge=0),
            ValenceAndCharge(valence=6, charge=0),
            ValenceAndCharge(valence=8, charge=0),
        ],
    ),
    "Cs": (
        [ValenceAndCharge(valence=1, charge=0)],
        [],
    ),
    "Ba": ([ValenceAndCharge(valence=2, charge=0)], []),
    "La": ([ValenceAndCharge(valence=3, charge=0)], []),
    "Ce": (
        [ValenceAndCharge(valence=3, charge=0)],
        [ValenceAndCharge(valence=4, charge=0)],
    ),
    "Pr": (
        [ValenceAndCharge(valence=3, charge=0)],
        [ValenceAndCharge(valence=4, charge=0)],
    ),
    "Nd": ([ValenceAndCharge(valence=3, charge=0)], []),
    "Pm": ([ValenceAndCharge(valence=3, charge=0)], []),
    "Sm": ([ValenceAndCharge(valence=3, charge=0)], []),
    "Eu": (
        [ValenceAndCharge(valence=3, charge=0)],
        [ValenceAndCharge(valence=2, charge=0)],
    ),
    "Gd": ([ValenceAndCharge(valence=3, charge=0)], []),
    "Tb": (
        [ValenceAndCharge(valence=3, charge=0)],
        [ValenceAndCharge(valence=4, charge=0)],
    ),
    "Dy": ([ValenceAndCharge(valence=3, charge=0)], []),
    "Ho": ([ValenceAndCharge(valence=3, charge=0)], []),
    "Er": ([ValenceAndCharge(valence=3, charge=0)], []),
    "Tm": ([ValenceAndCharge(valence=3, charge=0)], []),
    "Yb": (
        [ValenceAndCharge(valence=3, charge=0)],
        [ValenceAndCharge(valence=2, charge=0)],
    ),
    "Lu": ([ValenceAndCharge(valence=3, charge=0)], []),
    "Hf": ([ValenceAndCharge(valence=4, charge=0)], []),
    "Ta": ([ValenceAndCharge(valence=5, charge=0)], []),
    "W": (
        [ValenceAndCharge(valence=6, charge=0), ValenceAndCharge(valence=4, charge=0)],
        [
            ValenceAndCharge(valence=5, charge=0),
            ValenceAndCharge(valence=3, charge=0),
            ValenceAndCharge(valence=2, charge=0),
        ],
    ),
    "Re": (
        [
            ValenceAndCharge(valence=5, charge=0),
            ValenceAndCharge(valence=4, charge=0),
            ValenceAndCharge(valence=3, charge=0),
        ],
        [
            ValenceAndCharge(valence=7, charge=0),
            ValenceAndCharge(valence=6, charge=0),
            ValenceAndCharge(valence=2, charge=0),
            ValenceAndCharge(valence=1, charge=0),
            ValenceAndCharge(valence=0, charge=0),
        ],
    ),
    "Os": (
        [ValenceAndCharge(valence=4, charge=0)],
        [
            ValenceAndCharge(valence=8, charge=0),
            ValenceAndCharge(valence=6, charge=0),
            ValenceAndCharge(valence=2, charge=0),
        ],
    ),
    "Ir": (
        [ValenceAndCharge(valence=4, charge=0), ValenceAndCharge(valence=3, charge=0)],
        [ValenceAndCharge(valence=6, charge=0), ValenceAndCharge(valence=4, charge=0)],
    ),
    "Pt": (
        [ValenceAndCharge(valence=2, charge=0)],
        [ValenceAndCharge(valence=4, charge=0)],
    ),
    "Au": (
        [ValenceAndCharge(valence=3, charge=0)],
        [ValenceAndCharge(valence=1, charge=0)],
    ),
    "Hg": (
        [ValenceAndCharge(valence=2, charge=0)],
        [ValenceAndCharge(valence=1, charge=0)],
    ),
    "Tl": (
        [ValenceAndCharge(valence=3, charge=0)],
        [ValenceAndCharge(valence=1, charge=0)],
    ),
    "Pb": (
        [ValenceAndCharge(valence=4, charge=0)],
        [ValenceAndCharge(valence=2, charge=0)],
    ),
    "Bi": (
        [ValenceAndCharge(valence=3, charge=0), ValenceAndCharge(valence=1, charge=0)],
        [ValenceAndCharge(valence=5, charge=0)],
    ),
    "Po": (
        [ValenceAndCharge(valence=4, charge=0)],
        [ValenceAndCharge(valence=2, charge=0)],
    ),
    "At": (
        [ValenceAndCharge(valence=1, charge=0)],
        [
            ValenceAndCharge(valence=5, charge=0),
            ValenceAndCharge(valence=3, charge=0),
            ValenceAndCharge(valence=7, charge=0),
        ],
    ),
    "Rn": (
        [ValenceAndCharge(valence=0, charge=0)],
        [ValenceAndCharge(valence=2, charge=0)],
    ),
    "Fr": ([ValenceAndCharge(valence=1, charge=0)], []),
    "Ra": ([ValenceAndCharge(valence=2, charge=0)], []),
    "Ac": ([ValenceAndCharge(valence=3, charge=0)], []),
    "Th": ([ValenceAndCharge(valence=4, charge=0)], []),
    "Pa": (
        [ValenceAndCharge(valence=5, charge=0)],
        [ValenceAndCharge(valence=4, charge=0)],
    ),
    "U": (
        [ValenceAndCharge(valence=6, charge=0)],
        [
            ValenceAndCharge(valence=5, charge=0),
            ValenceAndCharge(valence=4, charge=0),
            ValenceAndCharge(valence=3, charge=0),
        ],
    ),
    "Np": (
        [ValenceAndCharge(valence=7, charge=0)],
        [
            ValenceAndCharge(valence=6, charge=0),
            ValenceAndCharge(valence=5, charge=0),
            ValenceAndCharge(valence=4, charge=0),
            ValenceAndCharge(valence=3, charge=0),
        ],
    ),
    "Pu": (
        [ValenceAndCharge(valence=7, charge=0), ValenceAndCharge(valence=4, charge=0)],
        [
            ValenceAndCharge(valence=6, charge=0),
            ValenceAndCharge(valence=5, charge=0),
            ValenceAndCharge(valence=3, charge=0),
        ],
    ),
    "Am": (
        [ValenceAndCharge(valence=3, charge=0)],
        [ValenceAndCharge(valence=5, charge=0), ValenceAndCharge(valence=4, charge=0)],
    ),
    "Cm": (
        [
            ValenceAndCharge(valence=6, charge=0),
            ValenceAndCharge(valence=5, charge=0),
            ValenceAndCharge(valence=3, charge=0),
        ],
        [],
    ),
    "Bk": (
        [ValenceAndCharge(valence=3, charge=0)],
        [ValenceAndCharge(valence=4, charge=0)],
    ),
    "Cf": ([ValenceAndCharge(valence=3, charge=0)], []),
    "Es": ([ValenceAndCharge(valence=3, charge=0)], []),
    "Fm": ([ValenceAndCharge(valence=3, charge=0)], []),
    "Md": ([ValenceAndCharge(valence=3, charge=0)], []),
    "No": (
        [ValenceAndCharge(valence=3, charge=0)],
        [ValenceAndCharge(valence=2, charge=0)],
    ),
    "Lr": ([ValenceAndCharge(valence=3, charge=0)], []),
    "Rf": ([ValenceAndCharge(valence=4, charge=0)], []),
    "Db": ([ValenceAndCharge(valence=5, charge=0)], []),
    "Sg": ([ValenceAndCharge(valence=6, charge=0)], []),
    "Bh": ([ValenceAndCharge(valence=7, charge=0)], []),
    "Hs": ([ValenceAndCharge(valence=8, charge=0)], []),
    "Mt": ([ValenceAndCharge(valence=8, charge=0)], []),
    "Ds": ([ValenceAndCharge(valence=8, charge=0)], []),
    "Rg": ([ValenceAndCharge(valence=8, charge=0)], []),
    "Cn": ([ValenceAndCharge(valence=2, charge=0)], []),
    "Nh": ([ValenceAndCharge(valence=3, charge=0)], []),
    "Fl": ([ValenceAndCharge(valence=4, charge=0)], []),
    "Mc": ([ValenceAndCharge(valence=3, charge=0)], []),
    "Lv": ([ValenceAndCharge(valence=4, charge=0)], []),
    "Ts": ([ValenceAndCharge(valence=7, charge=0)], []),
    "Og": ([ValenceAndCharge(valence=0, charge=0)], []),
    "+": ([ValenceAndCharge(valence=1, charge=1)], []),
    "-": ([ValenceAndCharge(valence=1, charge=-1)], []),
}



import typing as T
from abc import ABC, abstractmethod
from copy import deepcopy
import torch
from torchmetrics import Metric, MeanMetric
from torchmetrics.functional.retrieval.hit_rate import retrieval_hit_rate

from massspecgym.models.base import MassSpecGymModel, Stage
from massspecgym.simulation_utils.misc_utils import safelog
from massspecgym.simulation_utils.spec_utils import sparse_cosine_distance, \
    get_ints_transform_func, get_ints_untransform_func, batched_l1_normalize, \
    sparse_jensen_shannon_similarity
from massspecgym.simulation_utils.nn_utils import build_lr_scheduler
from massspecgym.utils import batch_ptr_to_batch_idx
from torch_geometric.utils import unbatch

def get_cos_sim_fn(transform_fn, mz_max, mz_bin_res):

    def cos_sim_fn(
        true_mzs: torch.Tensor, 
        true_logprobs: torch.Tensor,
        true_batch_idxs: torch.Tensor,
        pred_mzs: torch.Tensor,
        pred_logprobs: torch.Tensor,
        pred_batch_idxs: torch.Tensor
    ) -> torch.Tensor:

        true_logprobs = transform_fn(true_logprobs, true_batch_idxs)
        pred_logprobs = transform_fn(pred_logprobs, pred_batch_idxs)
        cos_sims = 1.-sparse_cosine_distance(
            pred_mzs=pred_mzs,
            pred_logprobs=pred_logprobs,
            pred_batch_idxs=pred_batch_idxs,
            true_mzs=true_mzs,
            true_logprobs=true_logprobs,
            true_batch_idxs=true_batch_idxs,
            mz_max=mz_max,
            mz_bin_res=mz_bin_res
        )
        return cos_sims

    return deepcopy(cos_sim_fn)

def get_js_sim_fn(transform_fn, mz_max, mz_bin_res):

    def js_sim_fn(
        true_mzs: torch.Tensor, 
        true_logprobs: torch.Tensor,
        true_batch_idxs: torch.Tensor,
        pred_mzs: torch.Tensor,
        pred_logprobs: torch.Tensor,
        pred_batch_idxs: torch.Tensor
    ) -> torch.Tensor:

        true_logprobs = transform_fn(true_logprobs, true_batch_idxs)
        pred_logprobs = transform_fn(pred_logprobs, pred_batch_idxs)
        js_sims = sparse_jensen_shannon_similarity(
            pred_mzs=pred_mzs,
            pred_logprobs=pred_logprobs,
            pred_batch_idxs=pred_batch_idxs,
            true_mzs=true_mzs,
            true_logprobs=true_logprobs,
            true_batch_idxs=true_batch_idxs,
            mz_max=mz_max,
            mz_bin_res=mz_bin_res
        )
        return js_sims

    return deepcopy(js_sim_fn)

class SimulationMassSpecGymModel(MassSpecGymModel, ABC):

    def __init__(self, **kwargs):

        super().__init__(**kwargs)
        # call setup functions
        self._setup_model()
        self._setup_loss_fn()
        self._setup_spec_fns()
        self._setup_metric_fns()

    @abstractmethod
    def _setup_model(self):

        pass

    def configure_optimizers(self):

        if self.hparams.optimizer_type == "adam":
            optimizer_cls = torch.optim.Adam
        elif self.hparams.optimizer_type == "adamw":
            optimizer_cls = torch.optim.AdamW
        elif self.hparams.optimizer_type == "sgd":
            optimizer_cls = torch.optim.SGD
        else:
            raise ValueError(f"Unknown optimizer {self.optimizer}")
        optimizer = optimizer_cls(
            self.parameters(), 
            lr=self.hparams.lr, 
            weight_decay=self.hparams.weight_decay
        )
        ret = {
            "optimizer": optimizer,
        }
        if self.hparams.lr_schedule:
            scheduler = build_lr_scheduler(
                optimizer=optimizer, 
                decay_rate=self.hparams.lr_decay_rate, 
                warmup_steps=self.hparams.lr_warmup_steps,
                decay_steps=self.hparams.lr_decay_steps,
            )
            ret["lr_scheduler"] = {
                "scheduler": scheduler,
                "frequency": 1,
                "interval": "step",
            }
        return ret

    def _setup_spec_fns(self):

        self.ints_transform_func = get_ints_transform_func(self.hparams.ints_transform)
        self.ints_untransform_func = get_ints_untransform_func(self.hparams.ints_transform)
        self.ints_normalize_func = batched_l1_normalize

    def _preproc_spec(self,spec_mzs,spec_ints,spec_batch_idxs):

        # transform
        spec_ints = spec_ints * 1000.
        spec_ints = self.ints_transform_func(spec_ints)
        # renormalize
        spec_ints = self.ints_normalize_func(
            spec_ints,
            spec_batch_idxs
        )
        spec_ints = safelog(spec_ints)
        return spec_mzs, spec_ints, spec_batch_idxs

    def _setup_loss_fn(self):

        def _loss_fn(
            true_mzs: torch.Tensor, 
            true_logprobs: torch.Tensor,
            true_batch_idxs: torch.Tensor,
            pred_mzs: torch.Tensor,
            pred_logprobs: torch.Tensor,
            pred_batch_idxs: torch.Tensor
        ):

            cos_dist = sparse_cosine_distance(
                true_mzs=true_mzs,
                true_logprobs=true_logprobs,
                true_batch_idxs=true_batch_idxs,
                pred_mzs=pred_mzs,
                pred_logprobs=pred_logprobs,
                pred_batch_idxs=pred_batch_idxs,
                mz_max=self.hparams.mz_max,
                mz_bin_res=self.hparams.mz_bin_res
            )
            return cos_dist

        self.loss_fn = _loss_fn

    def _setup_metric_fns(self):

        # untransformed
        def no_transform_fn(logprobs, batch_idxs):
            probs = self.ints_untransform_func(torch.exp(logprobs), batch_idxs)
            logprobs = safelog(self.ints_normalize_func(probs, batch_idxs))
            return logprobs
        self.cos_sim_fn = get_cos_sim_fn(
            deepcopy(no_transform_fn), 
            self.hparams.mz_max, 
            self.hparams.mz_bin_res
        )
        self.js_sim_fn = get_js_sim_fn(
            deepcopy(no_transform_fn), 
            self.hparams.mz_max, 
            self.hparams.mz_bin_res
        )
        # sqrt transform
        def sqrt_transform_fn(logprobs, batch_idxs):
            probs = self.ints_untransform_func(torch.exp(logprobs), batch_idxs)
            probs = probs**0.5
            logprobs = safelog(self.ints_normalize_func(probs, batch_idxs))
            return logprobs
        self.cos_sim_sqrt_fn = get_cos_sim_fn(
            deepcopy(sqrt_transform_fn), 
            self.hparams.mz_max,
            self.hparams.mz_bin_res
        )
        # obj transform
        obj_transform_fn = lambda x, y: x
        self.cos_sim_obj_fn = get_cos_sim_fn(
            deepcopy(obj_transform_fn), 
            self.hparams.mz_max,
            self.hparams.mz_bin_res
        )

    def forward(self, **kwargs) -> dict:

        return self.model.forward(**kwargs)

    def get_retrieval_outputs(
        self, 
        true_mzs,
        true_logprobs,
        true_batch_idxs,
        candidates_data,
        **kwargs
    ) -> dict:

        input_data = {}
        for k in candidates_data:
            input_data[k] = candidates_data[k]
        c_ptr = input_data["batch_ptr"]
        for k in ["precursor_mz", "adduct", "instrument_type", "collision_energy"]:
            if k in kwargs:
                assert not k in input_data, k
                input_data[k] = torch.repeat_interleave(kwargs[k],c_ptr)
        output_data = self.forward(**input_data)
        c_offset = torch.cumsum(c_ptr, dim=0)
        c_offset = torch.cat([
            torch.zeros((1,),dtype=c_offset.dtype,device=c_offset.device),
            c_offset
        ], dim=0)
        c_total = torch.sum(c_ptr)
        c_true_mzs, c_true_logprobs, c_true_batch_idxs = [], [], []
        # TODO: replace with unbatch?
        for idx in range(c_ptr.shape[0]):
            mask = true_batch_idxs==idx
            ptr = c_ptr[idx]
            c_true_mzs.append(true_mzs[mask].repeat(ptr))
            c_true_logprobs.append(true_logprobs[mask].repeat(ptr))
            c_true_batch_idxs.append(torch.repeat_interleave(torch.arange(ptr,device=c_ptr.device),mask.sum()) + c_offset[idx])
        c_true_mzs = torch.cat(c_true_mzs, dim=0)
        c_true_logprobs = torch.cat(c_true_logprobs, dim=0)
        c_true_batch_idxs = torch.cat(c_true_batch_idxs, dim=0)
        assert output_data["pred_batch_idxs"].max() == c_true_batch_idxs.max()
        c_scores = self.cos_sim_fn(
            true_mzs=c_true_mzs,
            true_logprobs=c_true_logprobs,
            true_batch_idxs=c_true_batch_idxs,
            pred_mzs=output_data["pred_mzs"],
            pred_logprobs=output_data["pred_logprobs"],
            pred_batch_idxs=output_data["pred_batch_idxs"]
        )
        c_labels = input_data["labels"]
        assert c_scores.shape[0] == c_labels.shape[0] == c_total, (c_scores.shape, c_labels.shape, c_total)
        # prepare dictionary
        ret_out_d = {}
        ret_out_d["retrieval_scores"] = c_scores
        ret_out_d["retrieval_labels"] = c_labels
        ret_out_d["retrieval_batch_ptr"] = c_ptr
        return ret_out_d

    def step(
        self, 
        batch: dict, 
        stage: Stage = Stage.NONE
    ) -> dict:

        true_mzs, true_logprobs, true_batch_idxs = self._preproc_spec(
            batch["spec_mzs"],
            batch["spec_ints"],
            batch["spec_batch_idxs"]
        )

        out_d = self.forward(
            **batch
        )
        pred_mzs = out_d["pred_mzs"]
        pred_logprobs = out_d["pred_logprobs"]
        pred_batch_idxs = out_d["pred_batch_idxs"]
        loss = self.loss_fn(
            true_mzs=true_mzs,
            true_logprobs=true_logprobs,
            true_batch_idxs=true_batch_idxs,
            pred_mzs=pred_mzs,
            pred_logprobs=pred_logprobs,
            pred_batch_idxs=pred_batch_idxs
        )
        mean_loss = torch.mean(loss)

        out_d = {
            "loss": mean_loss,
            "losses": loss, 
            "pred_mzs": pred_mzs, 
            "pred_logprobs": pred_logprobs, 
            "pred_batch_idxs": pred_batch_idxs,
            "true_mzs": true_mzs,
            "true_logprobs": true_logprobs,
            "true_batch_idxs": true_batch_idxs
        }

        if stage == Stage.TEST and self.hparams.do_retrieval:
            # retrieval
            assert "candidates_data" in batch, batch.keys()
            ret_out_d = self.get_retrieval_outputs(
                true_mzs=true_mzs,
                true_logprobs=true_logprobs,
                true_batch_idxs=true_batch_idxs,
                **batch
            )
            out_d.update(ret_out_d)

        return out_d

    def on_batch_end(
        self, outputs: T.Any, batch: dict, batch_idx: int, stage: Stage
    ) -> None:
        """
        Compute evaluation metrics for the retrieval model based on the batch and corresponding predictions.
        This method will be used in the `on_train_batch_end`, `on_validation_batch_end`, since `on_test_batch_end` is
        overridden below.
        """

        batch_size = torch.max(outputs["true_batch_idxs"])+1
        
        metric_vals = {}
        metric_vals["losses"] = outputs["losses"]

        # Log loss
        self.log(
            stage.to_pref() + "loss_step",
            outputs["loss"],
            batch_size=batch_size,
            sync_dist=True,
            prog_bar=True,
            on_step=True,
            on_epoch=False
        )

        if stage not in self.log_only_loss_at_stages:

            metric_vals.update(self.evaluate_similarity_step(
                pred_mzs=outputs["pred_mzs"],
                pred_logprobs=outputs["pred_logprobs"],
                pred_batch_idxs=outputs["pred_batch_idxs"],
                true_mzs=outputs["true_mzs"],
                true_logprobs=outputs["true_logprobs"],
                true_batch_idxs=outputs["true_batch_idxs"],
                stage=stage
            ))

            if stage == stage.TEST and self.df_test_path is not None:
                self._update_df_test(metric_vals)

    def on_test_batch_end(
        self, outputs: T.Any, batch: dict, batch_idx: int
    ) -> None:
        
        stage = Stage.TEST
        
        self.on_batch_end(
            outputs=outputs,
            batch=batch,
            batch_idx=batch_idx,
            stage=stage
        )

        if stage not in self.log_only_loss_at_stages and self.hparams.do_retrieval:
            
            metric_vals = {}
            metric_vals |= self.evaluate_retrieval_step(
                scores=outputs["retrieval_scores"],
                labels=outputs["retrieval_labels"],
                batch_ptr=outputs["retrieval_batch_ptr"],
                stage=stage
            )
            if self.df_test_path is not None:
                self._update_df_test(metric_vals)

    def evaluate_similarity_step(
        self,
        pred_mzs: torch.Tensor,
        pred_logprobs: torch.Tensor,
        pred_batch_idxs: torch.Tensor,
        true_mzs: torch.Tensor,
        true_logprobs: torch.Tensor,
        true_batch_idxs: torch.Tensor,
        stage: Stage
    ) -> dict:
        
        metric_vals = {}

        batch_size = torch.max(true_batch_idxs).item()+1
        assert "cos_sim" in self.hparams.sim_metrics, self.hparams.sim_metrics
        for metric_name in self.hparams.sim_metrics:
            metric_fn = getattr(self, metric_name+"_fn")
            metric = metric_fn(
                true_mzs=true_mzs,
                true_logprobs=true_logprobs,
                true_batch_idxs=true_batch_idxs,
                pred_mzs=pred_mzs,
                pred_logprobs=pred_logprobs,
                pred_batch_idxs=pred_batch_idxs
            )
            self._update_metric(
                name=stage.to_pref() + metric_name,
                metric_class=MeanMetric,
                update_args=(metric,),
                batch_size=batch_size,
                prog_bar=True,
                log=True,
                log_n_samples=False,
                bootstrap=(stage == Stage.TEST)
            )
            metric_vals[metric_name] = metric

        return metric_vals

    def evaluate_retrieval_step(
        self,
        scores: torch.Tensor,
        labels: torch.Tensor,
        batch_ptr: torch.Tensor,
        stage: Stage,
    ) -> dict:
        """
        Main evaluation method for the retrieval models. The retrieval step is evaluated by 
        computing the hit rate at different top-k values.

        Args:
            scores (torch.Tensor): Concatenated scores for all candidates for all samples in the 
                batch
            labels (torch.Tensor): Concatenated True/False labels for all candidates for all samples
                 in the batch
            batch_ptr (torch.Tensor): Number of each sample's candidates in the concatenated tensors
        """
        assert stage == Stage.TEST, stage

        # Initialize return dictionary to store metric values per sample
        metric_vals = {}

        # Evaluate hitrate at different top-k values
        indexes = batch_ptr_to_batch_idx(batch_ptr)
        scores = scores + 1e-5*torch.randn_like(scores)
        scores = unbatch(scores, indexes)
        labels = unbatch(labels, indexes)

        for at_k in self.hparams.at_ks:
            hit_rates = []
            for scores_sample, labels_sample in zip(scores, labels):
                hit_rates.append(retrieval_hit_rate(scores_sample, labels_sample, top_k=at_k))
            hit_rates = torch.tensor(hit_rates, device=batch_ptr.device)

            metric_name = f"{stage.to_pref()}hit_rate@{at_k}"
            self._update_metric(
                metric_name,
                MeanMetric,
                (hit_rates,),
                batch_size=batch_ptr.size(0),
                bootstrap=(stage == Stage.TEST)
            )
            metric_vals[metric_name] = hit_rates

        return metric_vals
    
    def on_test_epoch_end(self):
        self._save_df_test()


import pytest


def denovo_baseline_training():
    import pytorch_lightning as pl
    from pytorch_lightning import Trainer

    from massspecgym.data import MassSpecDataset, MassSpecDataModule
    from massspecgym.data.transforms import SpecTokenizer
    from massspecgym.models.de_novo import SmilesTransformer
    from massspecgym.models.tokenizers import SmilesBPETokenizer

    # Load dataset
    dataset = MassSpecDataset(
        spec_transform=SpecTokenizer(n_peaks=60),
        mol_transform=None
    )

    # Init data module
    data_module = MassSpecDataModule(
        dataset=dataset,
        batch_size=32
    )

    # Init model
    model = SmilesTransformer(
        input_dim=2,
        d_model=512,
        nhead=8,
        num_encoder_layers=4,
        num_decoder_layers=4,
        dropout=0.0,
        smiles_tokenizer=SmilesBPETokenizer(max_len=200),
        k_predictions=1,
        pre_norm=False,
        max_smiles_len=100,
        validate_only_loss=True
    )

    # Init trainer
    project = "MassSpecGymDeNovo"
    name = "SmilesTransformer_debug_overfitting"
    logger = pl.loggers.WandbLogger(
        project=project,
        name=name,
        tags=[],
        log_model=False,
        mode="disabled"
    )
    trainer = Trainer(
        accelerator="cpu", max_epochs=1, logger=logger, log_every_n_steps=1, check_val_every_n_epoch=50,
        limit_train_batches=3, limit_val_batches=3
    )

    # Train
    trainer.fit(model, datamodule=data_module)


def test_denovo_baseline_training():
    try:
        denovo_baseline_training()
    except Exception as e:
        pytest.fail(f"denovo_baseline_training() raised an exception: {e}")



import pytest
from rdkit import Chem
from rdkit.DataStructs import TanimotoSimilarity
from rdkit.Chem import AllChem
from tokenizers import Tokenizer
from typing import List
from massspecgym.models.tokenizers import SelfiesTokenizer, SmilesBPETokenizer
from massspecgym.utils import tanimoto_morgan_similarity


def assert_smiles_equal(smiles1: str, smiles2: str):
    """Assert that two SMILES strings represent the same molecule."""
    assert tanimoto_morgan_similarity(smiles1, smiles2) == 1.0


def assert_smiles_batch_equal(smiles_batch1: List[str], smiles_batch2: List[str]):
    """Assert that two lists of SMILES strings represent the same molecules."""
    assert len(smiles_batch1) == len(smiles_batch2)
    for smi1, smi2 in zip(smiles_batch1, smiles_batch2):
        assert_smiles_equal(smi1, smi2)


@pytest.fixture(scope="module")
def selfies_tokenizer():
    """Fixture for SelfiesTokenizer that ensures it's initialized once per module."""
    return SelfiesTokenizer(max_len=10)


@pytest.fixture(scope="module")
def bpe_tokenizer():
    """Fixture for SmilesBPETokenizer that ensures it's initialized once per module."""
    return SmilesBPETokenizer(max_len=10)


def test_selfies_encode_decode(selfies_tokenizer: Tokenizer):
    """Test encoding and decoding of a single SMILES string with SelfiesTokenizer."""
    smiles = "c1ccccc1"
    encoded = selfies_tokenizer.encode(smiles)
    assert isinstance(encoded.ids, list)
    assert len(encoded.ids) <= 10

    decoded = selfies_tokenizer.decode(encoded.ids)
    assert_smiles_equal(decoded, smiles)


def test_selfies_encode_decode_batch(selfies_tokenizer: Tokenizer):
    """Test batch encoding and decoding with SelfiesTokenizer."""
    smiles_batch = ["c1ccccc1", "O=C=O"]
    encoded_batch = selfies_tokenizer.encode_batch(smiles_batch)

    assert len(encoded_batch) == len(smiles_batch)
    for encoding in encoded_batch:
        assert isinstance(encoding.ids, list)
        assert len(encoding.ids) <= 10

    decoded_batch = selfies_tokenizer.decode_batch([encoding.ids for encoding in encoded_batch])
    assert_smiles_batch_equal(decoded_batch, smiles_batch)


def test_selfies_padding_truncation(selfies_tokenizer: Tokenizer):
    """Test padding and truncation with SelfiesTokenizer."""
    long_smiles = "C1CC(C=C1)C(=O)OCC2C(C(C(C(O2)OC3=C(C=C(C=C3)O)COC(=O)C4(C=CCCC4=O)O)OC(=O)C5CCC=C5)O)O"  # Longer SMILES string
    encoded = selfies_tokenizer.encode(long_smiles)
    assert len(encoded.ids) == 10  # Ensure it is truncated to max_len

    selfies_tokenizer.enable_padding(length=15)  # Increase padding length
    encoded = selfies_tokenizer.encode(long_smiles)
    assert len(encoded.ids) == 15  # Ensure it is padded to 15 tokens


def test_bpe_encode_decode(bpe_tokenizer: Tokenizer):
    """Test encoding and decoding of a single SMILES string with SmilesBPETokenizer."""
    smiles = "c1ccccc1"
    encoded = bpe_tokenizer.encode(smiles)
    assert isinstance(encoded.ids, list)
    assert len(encoded.ids) <= 10

    decoded = bpe_tokenizer.decode(encoded.ids)
    assert_smiles_equal(decoded, smiles)


def test_bpe_encode_decode_batch(bpe_tokenizer: Tokenizer):
    """Test batch encoding and decoding with SmilesBPETokenizer."""
    smiles_batch = ["c1ccccc1", "O=C=O"]
    encoded_batch = bpe_tokenizer.encode_batch(smiles_batch)

    assert len(encoded_batch) == len(smiles_batch)
    for encoding in encoded_batch:
        assert isinstance(encoding.ids, list)
        assert len(encoding.ids) <= 10

    decoded_batch = bpe_tokenizer.decode_batch([encoding.ids for encoding in encoded_batch])
    assert_smiles_batch_equal(decoded_batch, smiles_batch)


def test_bpe_padding_truncation(bpe_tokenizer: Tokenizer):
    """Test padding and truncation with SmilesBPETokenizer."""
    long_smiles = "C1CC(C=C1)C(=O)OCC2C(C(C(C(O2)OC3=C(C=C(C=C3)O)COC(=O)C4(C=CCCC4=O)O)OC(=O)C5CCC=C5)O)O"  # Longer SMILES string
    encoded = bpe_tokenizer.encode(long_smiles)
    assert len(encoded.ids) == 10  # Ensure it is truncated to max_len

    bpe_tokenizer.enable_padding(length=15)  # Increase padding length
    encoded = bpe_tokenizer.encode(long_smiles)
    assert len(encoded.ids) == 15  # Ensure it is padded to 15 tokens


def test_special_tokens(selfies_tokenizer: Tokenizer):
    """Test that special tokens are correctly added and handled in the tokenizer."""
    special_tokens = [selfies_tokenizer.pad_token, selfies_tokenizer.sos_token, selfies_tokenizer.eos_token]
    special_token_ids = [selfies_tokenizer.pad_token_id, selfies_tokenizer.sos_token_id, selfies_tokenizer.eos_token_id]

    for token, token_id in zip(special_tokens, special_token_ids):
        assert token_id == selfies_tokenizer.token_to_id(token)


def test_unpadding(selfies_tokenizer: Tokenizer):
    """Test unpadding functionality in the SelfiesTokenizer."""
    smiles = "c1ccccc1"
    encoded = selfies_tokenizer.encode(smiles)
    unpadded = [id for id in encoded.ids if id != selfies_tokenizer.pad_token_id]
    assert len(unpadded) <= len(encoded.ids)
    assert selfies_tokenizer.pad_token_id not in unpadded


from torch.utils.data import Subset
import numpy as np
import os
import yaml

from massspecgym.simulation_utils.misc_utils import deep_update

def load_config(template_fp, custom_fp):

    assert os.path.isfile(template_fp), template_fp
    if custom_fp:
        assert os.path.isfile(custom_fp), custom_fp
    with open(template_fp, "r") as template_file:
        config_d = yaml.load(template_file, Loader=yaml.FullLoader)
    # overwrite parts of the config
    if custom_fp:
        with open(custom_fp, "r") as custom_file:
            custom_d = yaml.load(custom_file, Loader=yaml.FullLoader)
        assert all([k in config_d for k in custom_d]), set(custom_d.keys()) - set(config_d.keys())
        config_d = deep_update(config_d, custom_d)
    return config_d

def get_split_ss(ds, split_type, subsample_frac=None):

    metadata = ds.metadata
    assert np.all(metadata.index == np.arange(metadata.shape[0]))
    if split_type == "benchmark":
        train_idxs = metadata[metadata["fold"]=="train"].index
        val_idxs = metadata[metadata["fold"]=="val"].index
        test_idxs = metadata[metadata["fold"]=="test"].index
    elif split_type == "all_inchikey":
        mol_ids = metadata["inchikey"].drop_duplicates()
        train_ids = mol_ids.sample(frac=0.8, replace=False, random_state=42)
        val_ids = mol_ids.drop(train_ids.index).sample(frac=0.5, replace=False, random_state=69)
        test_ids = mol_ids.drop(train_ids.index).drop(val_ids.index)
        train_idxs = metadata[metadata["inchikey"].isin(train_ids)].index
        val_idxs = metadata[metadata["inchikey"].isin(val_ids)].index
        test_idxs = metadata[metadata["inchikey"].isin(test_ids)].index
    elif split_type == "orbitrap_inchikey":
        mol_ids = metadata[metadata["instrument_type"]=="Orbitrap"]["inchikey"].drop_duplicates()
        train_ids = mol_ids.sample(frac=0.8, replace=False, random_state=42)
        val_ids = mol_ids.drop(train_ids.index).sample(frac=0.5, replace=False, random_state=69)
        test_ids = mol_ids.drop(train_ids.index).drop(val_ids.index)
        train_idxs = metadata[metadata["inchikey"].isin(train_ids)].index
        val_idxs = metadata[metadata["inchikey"].isin(val_ids)].index
        test_idxs = metadata[metadata["inchikey"].isin(test_ids)].index
    else:
        raise ValueError(f"split_type {split_type} not supported")
    if subsample_frac is not None:
        assert isinstance(subsample_frac, float)
        train_idxs = np.random.choice(
            train_idxs, 
            size=int(subsample_frac*len(train_idxs)),
            replace=False
        )
        val_idxs = np.random.choice(
            val_idxs, 
            size=int(subsample_frac*len(val_idxs)),
            replace=False
        )
        test_idxs = np.random.choice(
            test_idxs, 
            size=int(subsample_frac*len(test_idxs)),
            replace=False
        )
    # train_mol_ids = metadata.loc[train_idxs]["inchikey"].unique()
    # val_mol_ids = metadata.loc[val_idxs]["inchikey"].unique()
    # test_mol_ids = metadata.loc[test_idxs]["inchikey"].unique()
    # print(">>> Number of Spectra")
    # print(len(train_idxs), len(val_idxs), len(test_idxs))
    # print(">>> Number of Unique Molecules")
    # print(len(train_mol_ids), len(val_mol_ids), len(test_mol_ids))
    # get subsets
    train_ds = Subset(ds, train_idxs)
    val_ds = Subset(ds, val_idxs)
    test_ds = Subset(ds, test_idxs)
    return train_ds, val_ds, test_ds


import numpy as np
import matchms
import os
from rdkit.Chem import AllChem as Chem
import massspecgym.utils as utils


def test_pad_spectrum():
    spec = np.array(
        [
            [
                45.134823,
                45.13699,
                110.096245,
                130.064972,
                136.111862,
                277.16571,
                289.165924,
                307.177856,
                406.223083,
            ],
            [
                0.01082,
                0.01064,
                0.17184,
                0.1397,
                0.00874,
                1.0,
                0.52842,
                0.00793,
                0.43696,
            ],
        ],
        dtype=float,
    ).T

    i = 32
    spec_pad = utils.pad_spectrum(spec, i, pad_value=0)
    assert spec_pad.shape == (i, 2)
    assert (spec_pad[0, : spec.shape[0]] == spec[0]).all()
    assert (spec_pad[1, : spec.shape[0]] == spec[1]).all()
    assert (spec_pad == 0).sum() == 2 * (i - spec.shape[0])
    assert (utils.pad_spectrum(spec, spec.shape[0]) == spec).all()

    i = 5
    try:
        utils.pad_spectrum(spec, i)
    except ValueError:
        pass
    else:
        assert False


def test_morgan_fp():
    mol = Chem.MolFromSmiles("Cn1cnc2c1c(=O)n(C)c(=O)n2C")
    for i in [2048, 4096]:
        fp = utils.morgan_fp(mol, fp_size=i)
        assert fp.shape == (i,)
        assert fp.dtype == np.int32


def test_standardize_smiles():
    def asserts():
        assert utils.standardize_smiles("OCO") == "C(O)O"
        assert utils.standardize_smiles(["OCO", "CC"]) == ["C(O)O", "CC"]
    try:
        asserts()
    except ImportError:
        os.system("pip install git+https://github.com/boecker-lab/standardizeUtils@b415f1c51b49f6c5cd0e9c6ab89224c8ad657a35#egg=standardizeUtils")
        asserts()



from abc import ABC, abstractmethod
from typing import Optional
import torch
import torch.nn as nn
from torch_geometric.data import Batch

from massspecgym.simulation_utils.formula_embedder import get_embedder
from massspecgym.simulation_utils.nn_utils import SpecFFN, GNN, build_pool_module


class BaseModel(nn.Module, ABC):

    def __init__(
        self,
        metadata_insert_location: str,
        collision_energy_input_size: int,
        collision_energy_insert_size: int,
        adduct_input_size: int,
        adduct_insert_size: int,
        instrument_type_input_size: int,
        instrument_type_insert_size: int
    ):

        super().__init__()
        # init metadata
        self.metadata_insert_size = 0
        self.metadata_insert_location = metadata_insert_location
        self._check_metadata_insert_location(metadata_insert_location)
        self._collision_energy_init(
            collision_energy_input_size=collision_energy_input_size,
            collision_energy_insert_size=collision_energy_insert_size
        )
        self._adduct_init(
            adduct_input_size=adduct_input_size,
            adduct_insert_size=adduct_insert_size
        )
        self._instrument_type_init(
            instrument_type_input_size=instrument_type_input_size,
            instrument_type_insert_size=instrument_type_insert_size
        )

    @abstractmethod
    def _check_metadata_insert_location(self, metadata_insert_location: str):

        pass

    def _collision_energy_init(
        self,
        collision_energy_input_size: int,
        collision_energy_insert_size: int):

        # embedder
        if collision_energy_insert_size > 0:
            embedder = get_embedder("abs-sines", max_count_int=collision_energy_input_size)
            collision_energy_embedder = nn.Sequential(
                embedder,
                nn.Linear(embedder.num_dim,collision_energy_insert_size)
            )	
            self.collision_energy_embedder = collision_energy_embedder
            self.metadata_insert_size += collision_energy_insert_size

    def _adduct_init(
        self,
        adduct_input_size: int,
        adduct_insert_size: int):

        if adduct_insert_size > 0:
            self.adduct_embedder = nn.Embedding(adduct_input_size+1, adduct_insert_size)	
            self.metadata_insert_size += adduct_insert_size

    def _instrument_type_init(
        self,
        instrument_type_input_size: int,
        instrument_type_insert_size: int):

        if instrument_type_insert_size > 0:
            self.instrument_type_embedder = nn.Embedding(instrument_type_input_size+1, instrument_type_insert_size)
            self.metadata_insert_size += instrument_type_insert_size

    def embed_metadata(
        self, 
        collision_energy: Optional[torch.Tensor] = None, 
        adduct: Optional[torch.Tensor] = None, 
        instrument_type: Optional[torch.Tensor] = None
    ):

        metadata_embeds = []
        if hasattr(self,"collision_energy_embedder"):
            collision_energy_embed = self.collision_energy_embedder(collision_energy.reshape(-1,1))
            metadata_embeds.append(collision_energy_embed)
        if hasattr(self,"adduct_embedder"):
            adduct_embed = self.adduct_embedder(adduct)
            metadata_embeds.append(adduct_embed)
        if hasattr(self,"instrument_type_embedder"):
            instrument_type_embed = self.instrument_type_embedder(instrument_type)
            metadata_embeds.append(instrument_type_embed)
        metadata_embeds = torch.cat(metadata_embeds, dim=1)
        return metadata_embeds


class FPModel(BaseModel):

    def __init__(
        self,
        # input
        fps_input_size: int,
        metadata_insert_location: str,
        collision_energy_input_size: int,
        collision_energy_insert_size: int,
        adduct_input_size: int,
        adduct_insert_size: int,
        instrument_type_input_size: int,
        instrument_type_insert_size: int,
        # output
        mz_max: int,
        mz_bin_res: float,
        # model
        mlp_hidden_size: int,
        mlp_dropout: float,
        mlp_num_layers: int,
        mlp_use_residuals: bool,
        ff_prec_mz_offset: int,
        ff_bidirectional: bool,
        ff_output_map_size: int
    ):

        super().__init__(
            metadata_insert_location=metadata_insert_location,
            collision_energy_input_size=collision_energy_input_size,
            collision_energy_insert_size=collision_energy_insert_size,
            adduct_input_size=adduct_input_size,
            adduct_insert_size=adduct_insert_size,
            instrument_type_input_size=instrument_type_input_size,
            instrument_type_insert_size=instrument_type_insert_size
        )

        self.mlp_input_dim = fps_input_size + self.metadata_insert_size

        self.ffn = SpecFFN(
            input_size=self.mlp_input_dim,
            hidden_size=mlp_hidden_size,
            mz_max=mz_max,
            mz_bin_res=mz_bin_res,
            num_layers=mlp_num_layers,
            dropout=mlp_dropout,
            prec_mz_offset=ff_prec_mz_offset,
            bidirectional=ff_bidirectional,
            use_residuals=mlp_use_residuals,
            output_map_size=ff_output_map_size
        )

    def _check_metadata_insert_location(self, location):

        assert location in ["mlp","none"], f"metadata_insert_location={location} not supported"

    def forward(
        self,
        fps: torch.Tensor, 
        precursor_mz: torch.Tensor,
        collision_energy: torch.Tensor,
        adduct: torch.Tensor,
        instrument_type: torch.Tensor,
        **kwargs
    ):

        # embed metadata
        metadata_embeds = self.embed_metadata(
            collision_energy=collision_energy,
            adduct=adduct,
            instrument_type=instrument_type
        )

        # combine with fingerprint
        fh = torch.cat([fps, metadata_embeds], dim=1)

        # apply ffn
        pred_mzs, pred_logprobs, pred_batch_idxs = self.ffn(fh,precursor_mz)
        out_d = {
            "pred_mzs": pred_mzs,
            "pred_logprobs": pred_logprobs,
            "pred_batch_idxs": pred_batch_idxs
        }
        return out_d


class PrecOnlyModel(nn.Module):

    def __init__(self):

        super().__init__()
        self.dummy_params = nn.Parameter(torch.zeros((1,), dtype=torch.float32))

    def forward(
        self, 
        precursor_mz: torch.Tensor,
        **kwargs):

        pred_mzs = precursor_mz
        pred_logprobs = 0.*self.dummy_params + torch.zeros_like(pred_mzs)
        pred_batch_idxs = torch.arange(pred_mzs.shape[0],device=pred_mzs.device)

        out_d = {
            "pred_mzs": pred_mzs,
            "pred_logprobs": pred_logprobs,
            "pred_batch_idxs": pred_batch_idxs
        }
        return out_d


class GNNModel(BaseModel):

    def __init__(
        self,
        # input
        mol_node_feats_size: int,
        mol_edge_feats_size: int,
        metadata_insert_location: str,
        collision_energy_input_size: int,
        collision_energy_insert_size: int,
        adduct_input_size: int,
        adduct_insert_size: int,
        instrument_type_input_size: int,
        instrument_type_insert_size: int,
        # output
        mz_max: int,
        mz_bin_res: float,
        # model
        mol_hidden_size: int,
        mol_num_layers: int,
        mol_gnn_type: str,
        mol_normalization: str,
        mol_dropout: float,
        mol_pool_type: str,
        mlp_hidden_size: int,
        mlp_dropout: float,
        mlp_num_layers: int,
        mlp_use_residuals: bool,
        ff_prec_mz_offset: int,
        ff_bidirectional: bool,
        ff_output_map_size: int,
    ):
        super().__init__(
            metadata_insert_location=metadata_insert_location,
            collision_energy_input_size=collision_energy_input_size,
            collision_energy_insert_size=collision_energy_insert_size,
            adduct_input_size=adduct_input_size,
            adduct_insert_size=adduct_insert_size,
            instrument_type_input_size=instrument_type_input_size,
            instrument_type_insert_size=instrument_type_insert_size
        )

        # setup mol gnn
        self.mol_node_feats_size = mol_node_feats_size
        self.mol_edge_feats_size = mol_edge_feats_size
        if self.metadata_insert_location == "mol":
            self.mol_node_feats_size += self.metadata_insert_size
        mol_kwargs = {
            "node_feats_size": self.mol_node_feats_size,
            "edge_feats_size": self.mol_edge_feats_size,
            "hidden_size": mol_hidden_size,
            "num_layers": mol_num_layers,
            "gnn_type": mol_gnn_type,
            "dropout": mol_dropout,
            "normalization": mol_normalization,
        }
        # Mol GNN
        self.mol_embedder = GNN(**mol_kwargs)
        self.mol_pool_type = mol_pool_type
        self.mol_pool = build_pool_module(mol_pool_type,mol_hidden_size)

        # MLP input = GNN output
        self.mlp_input_dim = mol_hidden_size
        # metadata
        if self.metadata_insert_location == "mlp":
            self.mlp_input_dim += self.metadata_insert_size

        self.ffn = SpecFFN(
            input_size=self.mlp_input_dim,
            hidden_size=mlp_hidden_size,
            mz_max=mz_max,
            mz_bin_res=mz_bin_res,
            num_layers=mlp_num_layers,
            dropout=mlp_dropout,
            prec_mz_offset=ff_prec_mz_offset,
            bidirectional=ff_bidirectional,
            use_residuals=mlp_use_residuals,
            output_map_size=ff_output_map_size
        )

    def _check_metadata_insert_location(self, location):

        assert location in ["mlp","mol","none"], f"metadata_insert_location={location} not supported"

    def forward(
        self, 
        mol_pyg: Batch, 
        precursor_mz: torch.Tensor,
        collision_energy: torch.Tensor,
        adduct: torch.Tensor,
        instrument_type: torch.Tensor,
        **kwargs
    ):
        # mol features
        # mol_x: mol level node feature matrix
        # mol_edge_index: mol graph connectivity in COO format with shape [2, num_edges]
        # edge_attr: mol graph edge feature matrix with shape [num_edges, num_edge_features]
        # batch: sample idx repsect to current batch
        mol_x, mol_edge_index, mol_edge_attr, mol_batch = mol_pyg.x, mol_pyg.edge_index, mol_pyg.edge_attr, mol_pyg.batch

        # int_dtype = mol_edge_index.dtype
        batch_size = mol_batch[-1]+1

        # embed metadata
        metadata_embeds = self.embed_metadata(
            collision_energy=collision_energy,
            adduct=adduct,
            instrument_type=instrument_type
        )		

        # metadata embeddings at the node feature level
        if self.metadata_insert_location == "mol":
            mol_metadata_embeds = torch.repeat_interleave(mol_metadata_embeds,torch.unique(mol_batch,return_counts=True)[1],dim=0)
            mol_x = torch.cat([mol_x,mol_metadata_embeds],dim=1)
        
        # get per-atom embeddings
        mol_embed_gnn = self.mol_embedder(
            mol_x,
            mol_batch,
            mol_edge_index,
            mol_edge_attr
        )
        mol_embed_gnn_pool = self.mol_pool(mol_embed_gnn,mol_batch)
        ffn_input = mol_embed_gnn_pool

        if self.metadata_insert_location == "mlp":
            ffn_input = torch.cat([ffn_input,metadata_embeds],dim=1)

        # apply ffn
        pred_mzs, pred_logprobs, pred_batch_idxs = self.ffn(ffn_input,precursor_mz)
        out_d = {
            "pred_mzs": pred_mzs,
            "pred_logprobs": pred_logprobs,
            "pred_batch_idxs": pred_batch_idxs
        }
        return out_d




import numpy as np
import torch
import matchms
import matchms.filtering as ms_filters
from rdkit.Chem import AllChem as Chem
from typing import Optional
from abc import ABC, abstractmethod
import massspecgym.utils as utils
from massspecgym.definitions import CHEM_ELEMS


class SpecTransform(ABC):
    """
    Base class for spectrum transformations. Custom transformations should inherit from this class.
    The transformation consists of two consecutive steps:
        1. Apply a series of matchms filters to the input spectrum (method `matchms_transforms`).
        2. Convert the matchms spectrum to a torch tensor (method `matchms_to_torch`).

    This class provides a structure for creating custom spectrum transformation pipelines.
    """

    @abstractmethod
    def matchms_transforms(self, spec: matchms.Spectrum) -> matchms.Spectrum:
        """
        Apply a series of matchms filters to the input spectrum. Abstract method.

        Args:
            spec (matchms.Spectrum): Input spectrum to be filtered.

        Returns:
            matchms.Spectrum: Filtered spectrum after applying matchms filters.
        """
        pass

    @abstractmethod
    def matchms_to_torch(self, spec: matchms.Spectrum) -> torch.Tensor:
        """
        Convert a matchms spectrum to a torch tensor. Abstract method.

        Args:
            spec (matchms.Spectrum): Filtered spectrum to be converted to a torch tensor.

        Returns:
            torch.Tensor: Tensor representation of the input spectrum.
        """
        pass

    def __call__(self, spec: matchms.Spectrum) -> torch.Tensor:
        """
        Compose the matchms filters and the torch conversion.

        Args:
            spec (matchms.Spectrum): Input spectrum to be transformed.

        Returns:
            torch.Tensor: Tensor representation of the input spectrum after applying
                          the matchms filters and converting to torch tensor.
        """
        return self.matchms_to_torch(self.matchms_transforms(spec))


def default_matchms_transforms(
    spec: matchms.Spectrum,
    n_max_peaks: int = 60,
    mz_from: float = 10,
    mz_to: float = 1000,
) -> matchms.Spectrum:
    """
    Apply a default set of transformations to a spectrum using matchms filters.

    This function applies standard preprocessing steps to a spectrum, such as
    selecting peaks within a certain m/z range and limiting the number of peaks.

    Args:
        spec (matchms.Spectrum): The input spectrum to be transformed.
        n_max_peaks (int, optional): Maximum number of peaks to retain. Defaults to 60.
        mz_from (float, optional): Minimum m/z value to keep. Defaults to 10.
        mz_to (float, optional): Maximum m/z value to keep. Defaults to 1000.

    Returns:
        matchms.Spectrum: The spectrum after applying the default transformations.
    """
    spec = ms_filters.select_by_mz(spec, mz_from=mz_from, mz_to=mz_to)
   def __init__(
        self,
        max_mz: float = 1005,
        bin_width: float = 1,
        to_rel_intensities: bool = True,
    ) -> None:
        self.max_mz = max_mz
        self.bin_width = bin_width
        self.to_rel_intensities = to_rel_intensities
        if not (max_mz / bin_width).is_integer():
            raise ValueError("`max_mz` must be divisible by `bin_width`.")

    def matchms_transforms(self, spec: matchms.Spectrum) -> matchms.Spectrum:
        return default_matchms_transforms(spec, mz_to=self.max_mz, n_max_peaks=None)

    def matchms_to_torch(self, spec: matchms.Spectrum) -> torch.Tensor:
        """
        Bin the spectrum into a fixed number of bins.
        """
        binned_spec = self._bin_mass_spectrum(
            mzs=spec.peaks.mz,
            intensities=spec.peaks.intensities,
            max_mz=self.max_mz,
            bin_width=self.bin_width,
            to_rel_intensities=self.to_rel_intensities,
        )
        return torch.from_numpy(binned_spec)

    def _bin_mass_spectrum(
        self, mzs, intensities, max_mz, bin_width, to_rel_intensities=True
    ):
        # Calculate the number of bins
        num_bins = int(np.ceil(max_mz / bin_width))

        # Calculate the bin indices for each mass
        bin_indices = np.floor(mzs / bin_width).astype(int)

        # Filter out mzs that exceed max_mz
        valid_indices = bin_indices[mzs <= max_mz]
        valid_intensities = intensities[mzs <= max_mz]

        # Clip bin indices to ensure they are within the valid range
        valid_indices = np.clip(valid_indices, 0, num_bins - 1)

        # Initialize an array to store the binned intensities
        binned_intensities = np.zeros(num_bins)

        # Use np.add.at to sum intensities in the appropriate bins
        np.add.at(binned_intensities, valid_indices, valid_intensities)

        # Generate the bin edges for reference
        # bin_edges = np.arange(0, max_mz + bin_width, bin_width)

        # Normalize the intensities to relative intensities
        if to_rel_intensities:
            binned_intensities /= np.max(binned_intensities)

        return binned_intensities  # , bin_edges


class MolTransform(ABC):
    @abstractmethod
    def from_smiles(self, mol: str):
        """
        Convert a SMILES string to a tensor-like representation. Abstract method.
        """

    def __call__(self, mol: str):
        return self.from_smiles(mol)


class MolFingerprinter(MolTransform):
    def __init__(self, type: str = "morgan", fp_size: int = 2048, radius: int = 2):
        if type != "morgan":
            raise NotImplementedError(
                "Only Morgan fingerprints are implemented at the moment."
            )
        self.type = type
        self.fp_size = fp_size
        self.radius = radius

    def from_smiles(self, mol: str):
        mol = Chem.MolFromSmiles(mol)
        return utils.morgan_fp(
            mol, fp_size=self.fp_size, radius=self.radius, to_np=True
        )


class MolToInChIKey(MolTransform):
    def __init__(self, twod: bool = True) -> None:
        self.twod = twod

    def from_smiles(self, mol: str) -> str:
        mol = Chem.MolFromSmiles(mol)
        return utils.mol_to_inchi_key(mol, twod=self.twod)


class MolToFormulaVector(MolTransform):
    def __init__(self):
        self.element_index = {element: i for i, element in enumerate(CHEM_ELEMS)}

    def from_smiles(self, smiles: str):
        mol = Chem.MolFromSmiles(smiles)
        if mol is None:
            raise ValueError(f"Invalid SMILES string: {smiles}")

        # Add explicit hydrogens to the molecule
        mol = Chem.AddHs(mol)

        # Initialize a vector of zeros for the 118 elements
        formula_vector = np.zeros(118, dtype=np.int32)

        # Iterate over atoms in the molecule and count occurrences of each element
        for atom in mol.GetAtoms():
            symbol = atom.GetSymbol()
            if symbol in self.element_index:
                index = self.element_index[symbol]
                formula_vector[index] += 1
            else:
                raise ValueError(f"Element '{symbol}' not found in the list of 118 elements.")

        return formula_vector

    @staticmethod
    def num_elements():
        return len(CHEM_ELEMS)



from massspecgym.models.simulation.base import SimulationMassSpecGymModel
from massspecgym.simulation_utils.model_utils import PrecOnlyModel


class PrecOnlySimulationMassSpecGymModel(SimulationMassSpecGymModel):

    def _setup_model(self):

        self.model = PrecOnlyModel()



"""Global variables used across the package."""
import pathlib

# Dirs
MASSSPECGYM_ROOT_DIR = pathlib.Path(__file__).parent.absolute()
MASSSPECGYM_REPO_DIR = MASSSPECGYM_ROOT_DIR.parent
MASSSPECGYM_DATA_DIR = MASSSPECGYM_REPO_DIR / 'data'
MASSSPECGYM_TEST_RESULTS_DIR = MASSSPECGYM_DATA_DIR / 'test_results'
MASSSPECGYM_ASSETS_DIR = MASSSPECGYM_REPO_DIR / 'assets'

# Special tokens
PAD_TOKEN = "<pad>"
SOS_TOKEN = "<s>"
EOS_TOKEN = "</s>"
UNK_TOKEN = "<unk>"

# Chemistry
# List of all 118 elements (indexed by atomic number)
CHEM_ELEMS = [
    "H", "He", "Li", "Be", "B", "C", "N", "O", "F", "Ne", "Na", "Mg", "Al", "Si", "P", "S", "Cl", "Ar",
    "K", "Ca", "Sc", "Ti", "V", "Cr", "Mn", "Fe", "Co", "Ni", "Cu", "Zn", "Ga", "Ge", "As", "Se", "Br", "Kr",
    "Rb", "Sr", "Y", "Zr", "Nb", "Mo", "Tc", "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn", "Sb", "Te", "I", "Xe",
    "Cs", "Ba", "La", "Ce", "Pr", "Nd", "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb", "Lu",
    "Hf", "Ta", "W", "Re", "Os", "Ir", "Pt", "Au", "Hg", "Tl", "Pb", "Bi", "Po", "At", "Rn", "Fr", "Ra", "Ac",
    "Th", "Pa", "U", "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm", "Md", "No", "Lr", "Rf", "Db", "Sg", "Bh",
    "Hs", "Mt", "Ds", "Rg", "Cn", "Nh", "Fl", "Mc", "Lv", "Ts", "Og"
]


import typing as T
import collections
from enum import Enum
from abc import ABC, abstractmethod
from pathlib import Path

import torch
import pytorch_lightning as pl
from torchmetrics import Metric, SumMetric
import pandas as pd
from massspecgym.utils import ReturnScalarBootStrapper


class Stage(Enum):

    TRAIN = 'train'
    VAL = 'val'
    TEST = 'test'
    NONE = 'none'

    def to_pref(self) -> str:
        return f"{self.value}_" if self != Stage.NONE else ""


class MassSpecGymModel(pl.LightningModule, ABC):

    def __init__(
        self,
        lr: float = 1e-4,
        weight_decay: float = 0.0,
        log_only_loss_at_stages: T.Sequence[Stage | str] = (Stage.TRAIN,),
        no_mces_metrics_at_stages: T.Sequence[Stage | str] = (Stage.VAL,),
        log_mces_metrics_on_early_stopping: bool = False,
        bootstrap_metrics: bool = False,
        df_test_path: T.Optional[str | Path] = None,
        **kwargs
    ):
        super().__init__()
        self.save_hyperparameters()

        # Setup metring logging
        self.log_only_loss_at_stages = [
            Stage(s) if isinstance(s, str) else s for s in log_only_loss_at_stages
        ]

        self.no_mces_metrics_at_stages = [
            Stage(s) if isinstance(s, str) else s for s in no_mces_metrics_at_stages
        ]
        self.bootstrap_metrics = bootstrap_metrics

        # Init dictionary to store dataframe columns where rows correspond to samples
        # (for constructing test dataframe with predictions and metrics for each sample)
        self.df_test_path = Path(df_test_path) if df_test_path is not None else None
        self.df_test = collections.defaultdict(list)

    @abstractmethod
    def step(
        self, batch: dict, stage: Stage = Stage.NONE
    ) -> tuple[torch.Tensor, torch.Tensor]:
        raise NotImplementedError(
            "Method `step` must be implemented in the model-specific child class."
        )

    def training_step(
        self, batch: dict, batch_idx: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        return self.step(batch, stage=Stage.TRAIN)

    def validation_step(
        self, batch: dict, batch_idx: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        return self.step(batch, stage=Stage.VAL)

    def test_step(
        self, batch: dict, batch_idx: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        return self.step(batch, stage=Stage.TEST)

    @abstractmethod
    def on_batch_end(
        self, outputs: T.Any, batch: dict, batch_idx: int, stage: Stage
    ) -> None:
        """
        Method to be called at the end of each batch. This method should be implemented by a child,
        task-dedicated class and contain the evaluation necessary for the task.
        """
        raise NotImplementedError(
            "Method `on_batch_end` must be implemented in the task-specific child class."
        )

    def on_train_batch_end(self, *args, **kwargs):
        return self.on_batch_end(*args, **kwargs, stage=Stage.TRAIN)

    def on_validation_batch_end(self, *args, **kwargs):
        return self.on_batch_end(*args, **kwargs, stage=Stage.VAL)

    def on_test_batch_end(self, *args, **kwargs):
        return self.on_batch_end(*args, **kwargs, stage=Stage.TEST)

    def configure_optimizers(self):
        # use Adam as default
        return torch.optim.Adam(
            self.parameters(), lr=self.hparams.lr, weight_decay=self.hparams.weight_decay
        )

    def get_checkpoint_monitors(self) -> list[dict]:
        monitors = [
            {"monitor": f"{Stage.VAL.to_pref()}loss", "mode": "min", "early_stopping": True}
        ]
        return monitors

    def _update_metric(
        self,
        name: str,
        metric_class: type[Metric],
        update_args: T.Any,
        batch_size: T.Optional[int] = None,
        prog_bar: bool = False,
        metric_kwargs: T.Optional[dict] = None,
        log: bool = True,
        log_n_samples: bool = False,
        bootstrap: bool = False,
        num_bootstraps: int = 100
    ) -> None:
        """
        This method enables updating and logging metrics without instantiating them in advance in
        the __init__ method. The metrics are aggreated over batches and logged at the end of the
        epoch. If the metric does not exist yet, it is instantiated and added as an attribute to the
        model.
        """
        # Process arguments
        bootstrap = bootstrap and 'bootstrap_metrics' in self.hparams and self.hparams.bootstrap_metrics

        # Log total number of samples (useful for debugging)
        if log_n_samples:
            self._update_metric(
                name=name + "_n_samples",
                metric_class=SumMetric,
                update_args=(len(update_args[0]),),
                batch_size=1,
            )

        # Init metric if does not exits yet
        if hasattr(self, name):
            metric = getattr(self, name)
        else:
            if metric_kwargs is None:
                metric_kwargs = dict()
            metric = metric_class(**metric_kwargs)
            metric = metric.to(self.device)
            setattr(self, name, metric)

        # Update
        metric(*update_args)

        # Log
        if log:
            self.log(
                name,
                metric,
                prog_bar=prog_bar,
                batch_size=batch_size,
                on_step=False,
                on_epoch=True,
                add_dataloader_idx=False,
                metric_attribute=name  # Suggested by a torchmetrics error
            )

        # Bootstrap
        if bootstrap:
            def _bootsrapped_metric_class(**metric_kwargs):
                metric = metric_class(**metric_kwargs)
                return ReturnScalarBootStrapper(metric, std=True, num_bootstraps=num_bootstraps)

            self._update_metric(
                name=name + "_std",
                metric_class=_bootsrapped_metric_class,
                update_args=update_args,
                batch_size=batch_size,
                metric_kwargs=metric_kwargs,
            )

    def _update_df_test(self, dct: dict) -> None:
        for col, vals in dct.items():
            if isinstance(vals, torch.Tensor):
                vals = vals.tolist()
            self.df_test[col].extend(vals)

    def _save_df_test(self) -> None:
        # Save test data frame to disk
        if self.df_test_path is not None:
            df_test = pd.DataFrame(self.df_test)
            self.df_test_path.parent.mkdir(parents=True, exist_ok=True)
            df_test.to_pickle(self.df_test_path)